"""Utility functions related to :mod:`numpy`.

Note that these functions require the NumPy package.

See Also
--------
.. image:: https://badge.fury.io/py/numpy.svg
    :target: https://badge.fury.io/py/numpy

**NumPy** is the fundamental package needed for scientific computing with Python.
It provides:

* a powerful N-dimensional array object
* sophisticated (broadcasting) functions
* tools for integrating C/C++ and Fortran code
* useful linear algebra, Fourier transform, and random number capabilities


Index
-----
.. currentmodule:: nanoutils
.. autosummary::
{autosummary}

API
---
{autofunction}

"""

from __future__ import annotations

from math import factorial, nan
from typing import Iterable, Any
from itertools import combinations
from collections import abc

from .utils import raise_if, construct_api_doc
from .typing_utils import ArrayLike, DtypeLike

try:
    import numpy as np
    NUMPY_EX: None | Exception = None
except Exception as ex:
    NUMPY_EX = ex

from ._dtype_mapping import DTypeMapping, MutableDTypeMapping

__all__ = [
    'as_nd_array',
    'array_combinations',
    'fill_diagonal_blocks',
    'DTypeMapping',
    'MutableDTypeMapping',
]


@raise_if(NUMPY_EX)
def as_nd_array(
    array: Iterable[Any] | ArrayLike,
    dtype: DtypeLike,
    ndmin: int = 1,
    copy: bool = False,
) -> np.ndarray:
    """Construct a numpy array from an iterable or array-like object.

    Examples
    --------
    .. doctest:: python
        :skipif: NUMPY_EX is not None

        >>> from nanoutils import as_nd_array

        >>> as_nd_array(1, int)
        array([1])

        >>> as_nd_array([1, 2, 3, 4], int)
        array([1, 2, 3, 4])

        >>> iterator = iter([1, 2, 3, 4])
        >>> as_nd_array(iterator, int)
        array([1, 2, 3, 4])


    Parameters
    ----------
    array : :class:`~collections.abc.Iterable` or array-like
        An array-like object or an iterable consisting of scalars.
    dtype : :class:`type` or :class:`numpy.dtype`
        The data type of the to-be returned array.
    ndmin : :class:`int`
        The minimum dimensionality of the to-be returned array.
    copy : :class:`bool`
        If :data:`True`, always return a copy.

    Returns
    -------
    :class:`numpy.ndarray`
        A numpy array constructed from **value**.

    """
    try:
        return np.array(array, dtype=dtype, ndmin=ndmin, copy=copy)

    except TypeError as ex:
        if not isinstance(array, abc.Iterable):
            raise ex

        ret: np.ndarray = np.fromiter(array, dtype=dtype)
        ret.shape += (ndmin - ret.ndim) * (1,)
        return ret


@raise_if(NUMPY_EX)
def array_combinations(
    array: ArrayLike,
    r: int = 2,
    axis: int = -1,
) -> np.ndarray:
    r"""Construct an array with all :func:`~itertools.combinations` of **ar** along a use-specified axis.

    Examples
    --------
    .. doctest:: python
        :skipif: NUMPY_EX is not None

        >>> from nanoutils import array_combinations

        >>> array = [[1, 2, 3, 4],
        ...          [5, 6, 7, 8]]

        >>> array_combinations(array, r=2)
        array([[[1, 2],
                [5, 6]],
        <BLANKLINE>
               [[1, 3],
                [5, 7]],
        <BLANKLINE>
               [[1, 4],
                [5, 8]],
        <BLANKLINE>
               [[2, 3],
                [6, 7]],
        <BLANKLINE>
               [[2, 4],
                [6, 8]],
        <BLANKLINE>
               [[3, 4],
                [7, 8]]])

    Parameters
    ----------
    array : array-like, shape :math:`(m, \dotsc)`
        An :math:`n` dimensional array-like object.
    r : :class:`int`
        The length of each combination.
    axis : :class:`int`
        The axis used for constructing the combinations.

    Returns
    -------
    :class:`numpy.ndarray`, shape :math:`(k, \dotsc, r)`
        A :math:`n+1` dimensional array with all **ar** combinations (of length ``r``)
        along the user-specified **axis**.
        The variable :math:`k` herein represents the number of combinations:
        :math:`k = \dfrac{m! / r!}{(m-r)!}`.

    """  # noqa: E501
    ar = np.array(array, ndmin=1, copy=False)
    n = ar.shape[axis]

    # Identify the number of combinations
    try:
        combinations_len = int(factorial(n) / factorial(r) / factorial(n - r))
    except ValueError as ex:
        raise ValueError(f"'r' ({r!r}) expects a positive integer larger than or equal to the "
                         f"length of 'array' axis {axis!r} ({n!r})") from ex

    # Define the shape of the to-be returned array
    _shape = list(ar.shape)
    del _shape[axis]
    shape = (combinations_len,) + tuple(_shape) + (r,)

    # Create, fill and return the new array
    ret = np.empty(shape, dtype=ar.dtype)
    for i, idx in enumerate(combinations(range(n), r=r)):
        ret[i] = ar.take(idx, axis=axis)
    return ret


@raise_if(NUMPY_EX)
def fill_diagonal_blocks(
    array: np.ndarray,
    i: int,
    j: int,
    val: float = nan,
) -> None:
    """Fill diagonal blocks in **array** of size :math:`(i, j)`.

    The blocks are filled along the last 2 axes in **array**.
    Performs an inplace update of **array**.

    Examples
    --------
    .. doctest:: python
        :skipif: NUMPY_EX is not None

        >>> import numpy as np
        >>> from nanoutils import fill_diagonal_blocks

        >>> array = np.zeros((10, 15), dtype=int)
        >>> i = 2
        >>> j = 3

        >>> fill_diagonal_blocks(array, i, j, val=1)
        >>> print(array)
        [[1 1 1 0 0 0 0 0 0 0 0 0 0 0 0]
         [1 1 1 0 0 0 0 0 0 0 0 0 0 0 0]
         [0 0 0 1 1 1 0 0 0 0 0 0 0 0 0]
         [0 0 0 1 1 1 0 0 0 0 0 0 0 0 0]
         [0 0 0 0 0 0 1 1 1 0 0 0 0 0 0]
         [0 0 0 0 0 0 1 1 1 0 0 0 0 0 0]
         [0 0 0 0 0 0 0 0 0 1 1 1 0 0 0]
         [0 0 0 0 0 0 0 0 0 1 1 1 0 0 0]
         [0 0 0 0 0 0 0 0 0 0 0 0 1 1 1]
         [0 0 0 0 0 0 0 0 0 0 0 0 1 1 1]]

    Parameters
    ----------
    array : :class:`numpy.ndarray`
        A >= 2D NumPy array whose diagonal blocks are to be filled.
        Gets modified in-place.
    i : :class:`int`
        The size of the diagonal blocks along axis -2.
    j : :class:`int`
        The size of the diagonal blocks along axis -1.
    val : :class:`float`
        Value to be written on the diagonal.
        Its type must be compatible with that of the array **a**.


    :rtype: :data:`None`

    """
    if (j <= 0) or (i <= 0):
        raise ValueError(f"'i' and 'j' should be larger than 0; observed values: {i} & {j}")

    i0 = j0 = 0
    dim1 = array.shape[-2]
    while dim1 > i0:
        array[..., i0:i0+i, j0:j0+j] = val
        i0 += i
        j0 += j


__doc__ = construct_api_doc(globals())



"""A private module for containg the :class:`nanoutils.PartialPrepend` class.

Notes
-----
:class:`~nanoutils.PartialPrepend` should be imported from
either :mod:`nanoutils` or :mod:`nanoutils.utils`.

"""

from typing import List
from functools import partial

__all__: List[str] = []


class PartialPrepend(partial):
    """A :func:`~functools.partial` subclass where the ``*args`` are appended rather than prepended.

    Examples
    --------
    .. code:: python

        >>> from functools import partial
        >>> from nanoutils import PartialPrepend

        >>> func1 = partial(isinstance, 1)  # isinstance(1, ...)
        >>> func2 = PartialPrepend(isinstance, float)  # isinstance(..., float)

        >>> func1(int)  # isinstance(1, int)
        True

        >>> func2(1.0)  # isinstance(1.0, float)
        True

    """  # noqa: E501

    def __call__(self, *args, **keywords):
        """Call and return :attr:`~PartialReversed.func`."""
        keywords = {**self.keywords, **keywords}
        return self.func(*args, *self.args, **keywords)



"""A module for the :class:`DTypeMapping` and :class:`MutableDTypeMapping` classes."""

from __future__ import annotations

import sys
from collections.abc import Iterable, Iterator, Callable, Mapping
from typing import TypeVar, TYPE_CHECKING, Any, ClassVar

from .numpy_utils import NUMPY_EX
from .utils import raise_if, positional_only
from ._user_dict import _DictLike, UserMapping, MutableUserMapping, _SupportsKeysAndGetItem

if TYPE_CHECKING or NUMPY_EX is None:
    import numpy as np
    from numpy import dtype
else:
    dtype = "numpy.dtype"

if TYPE_CHECKING:
    import numpy.typing as npt
    from IPython.lib.pretty import RepresentationPrinter
    from typing_extensions import TypeGuard

__all__ = ["DTypeMapping", "MutableDTypeMapping"]

_T = TypeVar("_T")
_ST1 = TypeVar("_ST1", bound="DTypeMapping")
_ST2 = TypeVar("_ST2", bound="MutableDTypeMapping")


def _has_keys(obj: object) -> TypeGuard[_SupportsKeysAndGetItem[Any, Any]]:
    """Check if the passed obj has the :meth:`~dict.keys` method."""
    return callable(getattr(obj, "keys", None))


def _repr_helper(self: DTypeMapping, dtype_repr: Callable[[np.dtype[Any]], str]) -> str:
    """Helper function for :meth:`DTypeMapping.__repr__`."""
    cls = type(self)
    if len(self) == 0:
        return f"{cls.__name__}()"

    offset = max([len(i) for i in self], default=0)
    values = "\n".join(f"    {k:{offset}} = {dtype_repr(v)}," for k, v in self.items())
    return f"{cls.__name__}(\n{values}\n)"


class DTypeMapping(UserMapping[str, dtype]):
    """A mapping for creating structured dtypes.

    Examples
    --------
    .. code-block:: python

        >>> from nanoutils import DTypeMapping
        >>> import numpy as np

        >>> DType1 = DTypeMapping({"x": float, "y": float, "z": float, "symbol": (str, 2)})
        >>> print(DType1)
        DTypeMapping(
            x      = float64,
            y      = float64,
            z      = float64,
            symbol = <U2,
        )

        >>> DType1.x
        dtype('float64')

        >>> DType1.symbol
        dtype('<U2')

        >>> @DTypeMapping.from_type
        ... class DType2:
        ...     xyz = (float, 3)
        ...     symbol = (str, 2)
        ...     charge = np.int64

        >>> print(DType2)
        DTypeMapping(
            xyz    = ('<f8', (3,)),
            symbol = <U2,
            charge = int64,
        )

    """

    __slots__ = ("_dtype",)
    _SLOTS: ClassVar[frozenset[str]] = frozenset({"__weakref__", "_hash", "_dtype", "_dict"})

    @property
    def dtype(self) -> np.dtype[np.void]:
        """Get a structured dtype constructed from dtype mapping."""
        try:
            return self._dtype
        except AttributeError:
            pass
        self._dtype: np.dtype[np.void] = np.dtype(list(self.items()))
        return self._dtype

    @raise_if(NUMPY_EX)
    @positional_only
    def __init__(
        self,
        __iterable: None | _DictLike[str, npt.DTypeLike] = None,
        **fields: npt.DTypeLike,
    ) -> None:
        """Initialize the instance."""
        if __iterable is None:
            dct = {}
        elif _has_keys(__iterable):
            dct = {k: np.dtype(__iterable[k]) for k in __iterable.keys()}
        else:
            dct = {k: np.dtype(v) for k, v in __iterable}  # type: ignore[union-attr]
        dct.update({k: np.dtype(v) for k, v in fields.items()})
        super().__setattr__("_dict", dct)

    @classmethod
    @raise_if(NUMPY_EX)
    def from_type(cls: type[_ST1], type_obj: type) -> _ST1:
        """Construct a new dtype mapping from all public attributes of the decorated type object.

        Example
        -------
        .. code-block:: python

            >>> from nanoutils import DTypeMapping

            >>> @DTypeMapping.from_type
            ... class AtomsDType:
            ...     xyz = (float, 3)
            ...     symbol = (str, 2)
            ...     charge = np.int64

            >>> print(AtomsDType)
            DTypeMapping(
                xyz    = ('<f8', (3,)),
                symbol = <U2,
                charge = int64,
            )

        Parameters
        ----------
        type_obj : :class:`type`
            A type object or any object that supports :func:`vars`.

        """
        try:
            dct = vars(type_obj)
        except TypeError:
            raise TypeError(f"Expected a type object, got {type(type_obj).__name__!r}") from None
        return cls._reconstruct({k: np.dtype(v) for k, v in dct.items() if not k.startswith("_")})

    def __repr__(self) -> str:
        """Implement :func:`repr(self) <repr>`."""
        return _repr_helper(self, lambda dtype: f"numpy.{dtype!r}")

    def __str__(self) -> str:
        """Implement :class:`str(self) <str>`."""
        return _repr_helper(self, str)

    def _repr_pretty_(self, p: RepresentationPrinter, cycle: bool) -> None:
        """Entry point for the :mod:`IPython <IPython.lib.pretty>` pretty printer."""
        if cycle:
            cls = type(self)
            p.text(f"{cls.__name__}(...)")
        else:
            p.text(str(self))

    def __hash__(self) -> int:
        """Implement :func:`hash(self) <hash>`."""
        try:
            return self._hash
        except AttributeError:
            pass
        self._hash = hash(tuple(self.items()))
        return self._hash

    def __eq__(self, other: object) -> bool:
        """Implement :meth:`self == other <object.__eq__>`."""
        if not isinstance(other, DTypeMapping):
            return NotImplemented
        iterator = zip(self._dict.items(), other._dict.items())
        return all(i == j for i, j in iterator)

    @classmethod
    @raise_if(NUMPY_EX)
    def fromkeys(  # type: ignore[override]
        cls: type[_ST1],
        iterable: Iterable[str],
        value: npt.DTypeLike = None,
    ) -> _ST1:
        """Create a new dictionary with keys from iterable and values set to value."""
        value = np.dtype(value)
        dct = dict.fromkeys(iterable, value)
        return cls._reconstruct(dct)

    def __getattr__(self, name: str) -> np.dtype[Any]:
        """Implement :func:`getattr(self, name) <getattr>`."""
        try:
            return self[name]
        except KeyError:
            cls = type(self)
            raise AttributeError(f"{cls.__name__!r} object has no attribute {name!r}") from None

    def __setattr__(self, name: str, value: Any) -> None:
        """Implement :func:`setattr(self, name, value) <setattr>`."""
        cls = type(self)
        if name not in cls._SLOTS and name in self:
            raise AttributeError(f"{cls.__name__!r} object attribute {name!r} is read-only")
        return super().__setattr__(name, value)

    def __delattr__(self, name: str) -> None:
        """Implement :func:`delattr(self, name) <delattr>`."""
        cls = type(self)
        if name not in cls._SLOTS and name in self:
            raise AttributeError(f"{cls.__name__!r} object attribute {name!r} is read-only")
        return super().__delattr__(name)

    def __dir__(self) -> list[str]:
        """Implement :func:`dir(self) <dir>`."""
        return sorted(set(super().__dir__()) | self.keys())

    if sys.version_info < (3, 8):
        def __reversed__(self) -> Iterator[str]:
            """Implement :func:`reversed(self) <reversed>`."""
            return reversed(list(self))

    if sys.version_info >= (3, 9):
        def __or__(self: _ST1, other: Mapping[str, npt.DTypeLike]) -> _ST1:
            """Implement :meth:`self | other <object.__or__>`."""
            if not isinstance(other, Mapping):
                return NotImplemented

            cls = type(self)
            if isinstance(other, DTypeMapping):
                return cls._reconstruct(self._dict | other._dict)
            else:
                return cls._reconstruct(self._dict | {k: np.dtype(v) for k, v in other.items()})

        def __ror__(self: _ST1, other: Mapping[str, npt.DTypeLike]) -> _ST1:
            """Implement :meth:`other | self <object.__ror__>`."""
            if not isinstance(other, Mapping):
                return NotImplemented

            cls = type(self)
            if isinstance(other, DTypeMapping):
                return cls._reconstruct(other._dict | self._dict)
            else:
                return cls._reconstruct({k: np.dtype(v) for k, v in other.items()} | self._dict)


class MutableDTypeMapping(DTypeMapping, MutableUserMapping[str, dtype]):  # type: ignore[misc]
    """A mutable mapping for creating structured dtypes.

    Examples
    --------
    .. code-block:: python

        >>> from nanoutils import DTypeMapping
        >>> import numpy as np

        >>> DType1 = MutableDTypeMapping({"x": float, "y": float, "z": float, "symbol": (str, 2)})
        >>> print(DType1)
        MutableDTypeMapping(
            x      = float64,
            y      = float64,
            z      = float64,
            symbol = <U2,
        )

        >>> @MutableDTypeMapping.from_type
        ... class DType2:
        ...     xyz = (float, 3)
        ...     symbol = (str, 2)
        ...     charge = np.int64

        >>> print(DType2)
        MutableDTypeMapping(
            xyz    = ('<f8', (3,)),
            symbol = <U2,
            charge = int64,
        )

    """

    __slots__ = ()
    __hash__ = None  # type: ignore[assignment]

    @property
    def dtype(self) -> np.dtype[np.void]:
        """Get a structured dtype constructed from the mapping."""
        return np.dtype(list(self.items()))

    def __setitem__(self, key: str, value: npt.DTypeLike) -> None:
        """Implement :meth:`self[key] = value <object.__setitem__>`."""
        self._dict[key] = np.dtype(value)

    def __setattr__(self, name: str, value: npt.DTypeLike) -> None:
        """Implement :func:`setattr(self, name, value) <setattr>`."""
        cls = type(self)
        if name not in cls._SLOTS and name in self:
            self[name] = value
        else:
            return object.__setattr__(self, name, value)

    def __delattr__(self, name: str) -> None:
        """Implement :func:`delattr(self, name) <delattr>`."""
        cls = type(self)
        if name not in cls._SLOTS and name in self:
            del self[name]
        else:
            return object.__delattr__(self, name)

    @positional_only
    def update(self, __iterable: None | _DictLike = None, **fields: npt.DTypeLike) -> None:
        """Update the mapping from the passed mapping or iterable."""
        if __iterable is None:
            pass
        elif _has_keys(__iterable):
            self._dict.update({k: np.dtype(__iterable[k]) for k in __iterable.keys()})
        else:
            self._dict.update({k: np.dtype(v) for k, v in __iterable})  # type: ignore[union-attr]
        self._dict.update({k: np.dtype(v) for k, v in fields.items()})

    if sys.version_info >= (3, 9):
        def __ior__(self: _ST2, other: Mapping[str, npt.DTypeLike]) -> _ST2:
            """Implement :meth:`self |= other <object.__ior__>`."""
            if not isinstance(other, Mapping):
                return NotImplemented
            elif isinstance(other, DTypeMapping):
                self._dict |= other._dict
            else:
                self._dict |= {k: np.dtype(v) for k, v in other.items()}
            return self



"""An module with various :mod:`h5py`-related utilities.

Index
-----
.. currentmodule:: nanoutils
.. autosummary::
    RecursiveKeysView
    RecursiveValuesView
    RecursiveItemsView

API
---
.. autoclass:: RecursiveKeysView
.. autoclass:: RecursiveValuesView
.. autoclass:: RecursiveItemsView

"""

from __future__ import annotations

import sys
import abc
from collections import Counter
from collections.abc import Generator, MappingView, Iterator
from typing import NoReturn

if sys.version_info >= (3, 9):
    from collections.abc import KeysView, ValuesView, ItemsView
else:
    from typing import KeysView, ValuesView, ItemsView

from .utils import raise_if, construct_api_doc, VersionInfo

try:
    import h5py
    from h5py import Dataset as H5PyDataset
    H5PY_EX: None | Exception = None
    H5PY_VERSION = VersionInfo._make(h5py.version.version_tuple[:3])
except Exception as ex:
    H5PY_EX = ex
    H5PY_VERSION = VersionInfo(0, 0, 0)
    H5PyDataset = "h5py.Dataset"

__all__ = [
    'recursive_keys',
    'recursive_values',
    'recursive_items',
    'RecursiveKeysView',
    'RecursiveValuesView',
    'RecursiveItemsView',
]


class _RecursiveMappingView(MappingView, metaclass=abc.ABCMeta):
    """Base class for recursive group or file views."""

    __slots__ = ("__weakref__",)

    _mapping: h5py.Group
    __hash__ = None  # type: ignore[assignment]

    @raise_if(H5PY_EX)
    def __init__(self, f: h5py.Group) -> None:
        if not isinstance(f, h5py.Group):
            raise TypeError("Expected a h5py Group or File")
        super().__init__(f)

    @property
    def mapping(self) -> h5py.Group:
        """:class:`h5py.Group`: Get the underlying mapping."""
        return self._mapping

    @classmethod
    def _iter_dfs(
        cls,
        group: h5py.Group,
        reverse: bool = False,
    ) -> Generator[tuple[str, h5py.Dataset], None, None]:
        """Recursively iterate through **group** in depth-first order."""
        iterable = group.values() if not reverse else reversed(group.values())
        for v in iterable:
            if isinstance(v, h5py.Group):
                yield from cls._iter_dfs(v, reverse)
            else:
                yield v.name, v

    def __len__(self) -> int:
        """Implement :func:`len(self)<len>`."""
        i = 0
        for i, _ in enumerate(self, 1):
            pass
        return i

    def __contains__(self, key: object) -> bool:
        """Implement :meth:`key in self<object.__contains__>`."""
        return any((key == i) for i in self)

    @abc.abstractmethod
    def __iter__(self) -> Iterator[object]:
        """Implement :func:`iter(self)<iter>`."""
        raise NotImplementedError("Trying to call an abstract method")

    if H5PY_VERSION >= (3, 5, 0):
        @abc.abstractmethod
        def __reversed__(self) -> Iterator[object]:
            """Implement :func:`reversed(self)<reversed>`.

            Note
            ----
            This feature requires h5py >= 3.5.

            """
            raise NotImplementedError("Trying to call an abstract method")
    else:
        def __reversed__(self) -> NoReturn:
            """Implement :func:`reversed(self)<reversed>`.

            Warning
            -------
            This feature requires h5py >= 3.5.

            """
            raise TypeError("`reversed` requires h5py >= 3.5.0") from H5PY_EX


class RecursiveKeysView(_RecursiveMappingView, KeysView[str]):
    """Create a recursive view of all dataset :attr:`names<h5py.Dataset.name>`.

    Examples
    --------
    .. testsetup:: python

        >>> import os
        >>> filename = os.path.join('tests', 'test_files', '_test.hdf5')
        >>> if os.path.isfile(filename):
        ...     os.remove(filename)

    .. doctest:: python
        :skipif: HDF5_EX is not None

        >>> import h5py
        >>> from nanoutils import RecursiveKeysView

        >>> filename: str = ...  # doctest: +SKIP

        >>> with h5py.File(filename, 'a') as f:
        ...     a = f.create_group('a')
        ...     b = f['a'].create_group('b')
        ...
        ...     dset1 = f.create_dataset('dset1', (10,), dtype=float)
        ...     dset2 = f['a'].create_dataset('dset2', (10,), dtype=float)
        ...     dset3 = f['a']['b'].create_dataset('dset3', (10,), dtype=float)

        >>> with h5py.File(filename, 'r') as f:
        ...     for key in RecursiveKeysView(f):
        ...         print(repr(key))
        '/a/b/dset3'
        '/a/dset2'
        '/dset1'


    .. testcleanup:: python

        >>> if os.path.isfile(filename):
        ...     os.remove(filename)

    Parameters
    ----------
    f : :class:`h5py.Group`
        The to-be iterated h5py group of file.

    Returns
    -------
    :class:`KeysView[str]<collections.abc.KeysView>`
        A recursive view of all dataset names within the passed group or file.

    """

    __slots__ = ()

    def __iter__(self) -> Generator[str, None, None]:
        """Implement :func:`iter(self)<iter>`."""
        for k, _ in self._iter_dfs(self._mapping):
            yield k

    if H5PY_VERSION >= (3, 5, 0):
        def __reversed__(self) -> Generator[str, None, None]:
            """Implement :func:`reversed(self)<reversed>`.

            Note
            ----
            This feature requires h5py >= 3.5.

            """
            for k, _ in self._iter_dfs(self._mapping, reverse=True):
                yield k


class RecursiveValuesView(_RecursiveMappingView, ValuesView[H5PyDataset]):
    """Create a recursive view of all :class:`<Datasets>h5py.Dataset`.

    Examples
    --------
    .. testsetup:: python

        >>> import os
        >>> filename = os.path.join('tests', 'test_files', '_test.hdf5')
        >>> if os.path.isfile(filename):
        ...     os.remove(filename)

    .. doctest:: python
        :skipif: HDF5_EX is not None

        >>> import h5py
        >>> from nanoutils import RecursiveValuesView

        >>> filename: str = ...  # doctest: +SKIP

        >>> with h5py.File(filename, 'a') as f:
        ...     a = f.create_group('a')
        ...     b = f['a'].create_group('b')
        ...
        ...     dset1 = f.create_dataset('dset1', (10,), dtype=float)
        ...     dset2 = f['a'].create_dataset('dset2', (10,), dtype=float)
        ...     dset3 = f['a']['b'].create_dataset('dset3', (10,), dtype=float)

        >>> with h5py.File(filename, 'r') as f:
        ...     for value in RecursiveValuesView(f):
        ...         print(value)
        <HDF5 dataset "dset3": shape (10,), type "<f8">
        <HDF5 dataset "dset2": shape (10,), type "<f8">
        <HDF5 dataset "dset1": shape (10,), type "<f8">


    .. testcleanup:: python

        >>> if os.path.isfile(filename):
        ...     os.remove(filename)

    Parameters
    ----------
    f : :class:`h5py.Group`
        The to-be iterated h5py group of file.

    Returns
    -------
    :class:`ValuesView[h5py.Dataset]<collections.abc.ValuesView>`
        A recursive view of all datasets within the passed group or file.

    """

    __slots__ = ()

    def __eq__(self, other: object) -> bool:
        """Implement :meth:`self == other<object.__eq__>`."""
        cls = type(self)
        if not isinstance(other, cls):
            return NotImplemented
        return Counter(self) == Counter(other)

    def __iter__(self) -> Generator[h5py.Dataset, None, None]:
        """Implement :func:`iter(self)<iter>`."""
        for _, v in self._iter_dfs(self._mapping):
            yield v

    if H5PY_VERSION >= (3, 5, 0):
        def __reversed__(self) -> Generator[h5py.Dataset, None, None]:
            """Implement :func:`reversed(self)<reversed>`.

            Note
            ----
            This feature requires h5py >= 3.5.

            """
            for _, v in self._iter_dfs(self._mapping, reverse=True):
                yield v


class RecursiveItemsView(_RecursiveMappingView, ItemsView[str, H5PyDataset]):
    """Create a recursive view of all :attr:`~h5py.Dataset.name`/:attr:`~h5py.Dataset` pairs.

    Examples
    --------
    .. testsetup:: python

        >>> import os
        >>> filename = os.path.join('tests', 'test_files', '_test.hdf5')
        >>> if os.path.isfile(filename):
        ...     os.remove(filename)

    .. doctest:: python
        :skipif: HDF5_EX is not None

        >>> import h5py
        >>> from nanoutils import RecursiveItemsView

        >>> filename: str = ...  # doctest: +SKIP

        >>> with h5py.File(filename, 'a') as f:
        ...     a = f.create_group('a')
        ...     b = f['a'].create_group('b')
        ...
        ...     dset1 = f.create_dataset('dset1', (10,), dtype=float)
        ...     dset2 = f['a'].create_dataset('dset2', (10,), dtype=float)
        ...     dset3 = f['a']['b'].create_dataset('dset3', (10,), dtype=float)

        >>> with h5py.File(filename, 'r') as f:
        ...     for items in RecursiveItemsView(f):
        ...         print(items)
        ('/a/b/dset3', <HDF5 dataset "dset3": shape (10,), type "<f8">)
        ('/a/dset2', <HDF5 dataset "dset2": shape (10,), type "<f8">)
        ('/dset1', <HDF5 dataset "dset1": shape (10,), type "<f8">)


    .. testcleanup:: python

        >>> if os.path.isfile(filename):
        ...     os.remove(filename)

    Parameters
    ----------
    f : :class:`h5py.Group`
        The to-be iterated h5py group of file.

    Returns
    -------
    :class:`ItemsView[str, h5py.Dataset]<collections.abc.ItemsView>`
        A recursive view of all name/dataset pairs within the passed group or file.

    """

    __slots__ = ()

    def __iter__(self) -> Generator[tuple[str, h5py.Dataset], None, None]:
        """Implement :func:`iter(self)<iter>`."""
        yield from self._iter_dfs(self._mapping)

    if H5PY_VERSION >= (3, 5, 0):
        def __reversed__(self) -> Generator[tuple[str, h5py.Dataset], None, None]:
            """Implement :func:`reversed(self)<reversed>`.

            Note
            ----
            This feature requires h5py >= 3.5.

            """
            yield from self._iter_dfs(self._mapping, reverse=True)


recursive_keys = RecursiveKeysView
recursive_items = RecursiveItemsView
recursive_values = RecursiveValuesView

__doc__ = construct_api_doc(globals())



"""A module with :mod:`schema`-related utility functions.

See Also
--------
.. image:: https://badge.fury.io/py/schema.svg
    :target: https://badge.fury.io/py/schema

**schema** is a library for validating Python data structures,
such as those obtained from config-files, forms,
external services or command-line parsing, converted from JSON/YAML
(or something else) to Python data-types.


Index
-----
.. currentmodule:: nanoutils
.. autosummary::
    {autosummary}

API
---
{autofunction}

"""

from __future__ import annotations

import inspect
import warnings
from typing import TypeVar, SupportsFloat, Callable, overload, Generic, Tuple
from numbers import Integral

from .typing_utils import Literal
from .utils import PartialPrepend, get_importable, construct_api_doc, get_func_name

__all__ = ['Default', 'Formatter', 'supports_float', 'supports_int',
           'isinstance_factory', 'issubclass_factory', 'import_factory']

_T = TypeVar('_T')


@overload
def supports_float(value: SupportsFloat) -> Literal[True]:
    ...
@overload  # noqa: E302
def supports_float(value: object) -> bool:
    ...
def supports_float(value: object) -> bool:  # noqa: E302
    """Check if a float-like object has been passed (:class:`~typing.SupportsFloat`).

    Examples
    --------
    .. code:: python

        >>> from nanoutils import supports_float

        >>> supports_float(1.0)
        True

        >>> supports_float(1)
        True

        >>> supports_float('1.0')
        True

        >>> supports_float('not a float')
        False

    Parameters
    ----------
    value : :class:`object`
        The to-be evaluated object.

    Returns
    -------
    :class:`bool`
        Whether or not the passed **value** is float-like or not.

    """
    try:
        float(value)  # type: ignore
        return True
    except Exception:
        return False


@overload
def supports_int(value: int | Integral) -> Literal[True]:
    ...
@overload  # noqa: E302
def supports_int(value: object) -> bool:
    ...
def supports_int(value: object) -> bool:  # noqa: E302
    """Check if an int-like object has been passed (:class:`~typing.SupportsInt`).

    Examples
    --------
    .. code:: python

        >>> from nanoutils import supports_int

        >>> supports_int(1.0)
        True

        >>> supports_int(1.5)
        False

        >>> supports_int(1)
        True

        >>> supports_int('1')
        True

        >>> supports_int('not a int')
        False

    Parameters
    ----------
    value : :class:`object`
        The to-be evaluated object.

    Returns
    -------
    :class:`bool`
        Whether or not the passed **value** is int-like or not.

    """
    # floats that can be exactly represented by an integer are also fine
    try:
        int(value)  # type: ignore
        return float(value).is_integer()  # type: ignore
    except Exception:
        return False


class Default(Generic[_T]):
    """A validation class akin to the likes of :class:`schemas.Use`.

    Upon executing :meth:`Default.validate` returns the stored :attr:`~Default.value`.
    If :attr:`~.Default.call` is ``True`` and the value is a callable,
    then it is called before its return.

    Examples
    --------
    .. code:: python

        >>> from schema import Schema, And
        >>> from nanoutils import Default

        >>> schema1 = Schema(And(int, Default(True)))
        >>> schema1.validate(1)
        True

        >>> schema2 = Schema(And(int, Default(dict)))
        >>> schema2.validate(1)
        {}

        >>> schema3 = Schema(And(int, Default(dict, call=False)))
        >>> schema3.validate(1)
        <class 'dict'>


    Attributes
    ----------
    value : :class:`object`
        The to-be return value for when :meth:`Default.validate` is called.
        If :attr:`Default.call` is ``True`` then the value is called
        (if possible) before its return.

    call : :class:`bool`
        Whether to call :attr:`Default.value` before its return (if possible) or not.

    """

    value: _T
    call: bool

    def __init__(self, value: _T, call: bool = True) -> None:
        """Initialize an instance."""
        self.value = value
        self.call = call

    def __repr__(self) -> str:
        """Implement :code:`str(self)` and :code:`repr(self)`."""
        return f'{self.__class__.__name__}({self.value!r}, call={self.call!r})'

    def validate(self, *args: object, **kwargs: object) -> _T:
        r"""Validate the passed **data**.

        Parameters
        ----------
        \*args/\**kwargs
            Variadic (keyword) arguments to ensure signature compatibility.
            Supplied values will not be used.

        Returns
        -------
        :class:`object`
            Return :attr:`Default.value`.
            The to-be returned value will be called if
            it is a callable and :attr:`Default.call` is :data:`True`.

        """
        if self.call and callable(self.value):
            return self.value()  # type: ignore
        else:
            return self.value


class Formatter(str):
    r"""A :class:`str` subclass used for creating :mod:`schema` error messages.

    Examples
    --------
    .. code:: python

        >>> from nanoutils import Formatter

        >>> string = Formatter("{name}: {type} = {value}")
        >>> string.format(1)
        'value: int = 1'

    """

    def __init__(self, msg: str) -> None:
        """Initialize an instance."""
        self._msg: str = msg

    def __repr__(self) -> str:
        """Implement :code:`str(self)` and :code:`repr(self)`."""
        return f'{self.__class__.__name__}(msg={self._msg!r})'

    def format(self, obj: object) -> str:  # type: ignore
        r"""Return a formatted version of :attr:`Formatter._msg`.

        Parameters
        ----------
        obj : :class:`object`
            The to-be formatted object.

        Returns
        -------
        :class:`str`
            A formatted string.

        """
        try:
            name = self._msg.split("'", maxsplit=2)[1]
        except IndexError:
            name = ''
        name_ = name or 'value'

        try:
            return self._msg.format(name=name_, value=obj, type=obj.__class__.__name__)
        except Exception as ex:
            err = RuntimeWarning(ex)
            err.__cause__ = ex
            warnings.warn(err)
            return repr(obj)

    @property
    def __mod__(self):
        """Get :meth:`Formatter.format`."""
        return self.format


_PO = inspect.Parameter.POSITIONAL_ONLY
_POK = inspect.Parameter.POSITIONAL_OR_KEYWORD


def isinstance_factory(
    class_or_tuple: type | Tuple[type, ...],
    module: str = __name__,
) -> Callable[[object], bool]:
    """Return a function which checks if the passed object is an instance of **class_or_tuple**.

    Examples
    --------
    .. code:: python

        >>> from nanoutils import isinstance_factory

        >>> func = isinstance_factory(int)

        >>> func(1)  # isinstance(1, int)
        True

        >>> func(1.0)  # isinstance(1.0, int)
        False

    Parameters
    ----------
    class_or_tuple : :class:`type` or :data:`Tuple[type, ...]<typing.Tuple>`
        A type object or tuple of type objects.
    module : :class:`str`
        The :attr:`~definition.__module__` of the to-be returned function.

    Returns
    -------
    :data:`Callable[[object], bool]<typing.Callable>`
        A function which asserts the passed object is an instance of **class_or_tuple**.

    See Also
    --------
    :func:`isinstance`
        Return whether an object is an instance of a class or of a subclass thereof.

    """
    try:
        isinstance('bob', class_or_tuple)
    except TypeError as ex:
        raise TypeError("'class_or_tuple' expected a type or tuple of types; "
                        f"observed type: {class_or_tuple.__class__.__name__!r}") from ex

    if isinstance(class_or_tuple, type):
        cls_str = class_or_tuple.__name__
    else:
        cls_str = f'({", ".join(cls.__name__ for cls in class_or_tuple)})'

    ret = PartialPrepend(isinstance, class_or_tuple)
    ret.__name__ = ret.func.__name__  # type: ignore
    ret.__qualname__ = ret.func.__qualname__  # type: ignore
    ret.__module__ = module
    ret.__doc__ = f'Return :code:`isinstance(obj, {cls_str})`.'
    ret.__signature__ = inspect.Signature(  # type: ignore
        parameters=[inspect.Parameter('obj', _PO, annotation=object)],
        return_annotation=bool
    )
    return ret


def issubclass_factory(
    class_or_tuple: type | Tuple[type, ...],
    module: str = __name__,
) -> Callable[[type], bool]:
    """Return a function which checks if the passed class is a subclass of **class_or_tuple**.

    Examples
    --------
    .. code:: python

        >>> from nanoutils import issubclass_factory

        >>> func = issubclass_factory(int)

        >>> func(bool)  # issubclass(bool, int)
        True

        >>> func(float)  # issubclass(float, int)
        False

    Parameters
    ----------
    class_or_tuple : :class:`type` or :data:`Tuple[type, ...]<typing.Tuple>`
        A type object or tuple of type objects.
    module : :class:`str`
        The :attr:`~definition.__module__` of the to-be returned function.

    Returns
    -------
    :data:`Callable[[type], bool]<typing.Callable>`
        A function which asserts the passed type is a subclass of **class_or_tuple**.

    See Also
    --------
    :func:`issubclass`
        Return whether **cls** is a derived from another class or is the same class.

    """
    try:
        issubclass(int, class_or_tuple)
    except TypeError as ex:
        raise TypeError("'class_or_tuple' expected a type or tuple of types; "
                        f"observed type: {class_or_tuple.__class__.__name__!r}") from ex

    if isinstance(class_or_tuple, type):
        cls_str = class_or_tuple.__name__
    else:
        cls_str = f'({", ".join(cls.__name__ for cls in class_or_tuple)})'

    ret = PartialPrepend(issubclass, class_or_tuple)
    ret.__name__ = ret.func.__name__  # type: ignore
    ret.__qualname__ = ret.func.__qualname__  # type: ignore
    ret.__module__ = module
    ret.__doc__ = f'Return :code:`isinstance(obj, {cls_str})`.'
    ret.__signature__ = inspect.Signature(  # type: ignore
        parameters=[inspect.Parameter('cls', _PO, annotation=type)],
        return_annotation=bool
    )
    return ret


def import_factory(
    validate: Callable[[_T], bool],
    module: str = __name__,
) -> Callable[[str], _T]:
    """Return a function which calls :func:`nanoutils.get_importable` with the **validate** argument.

    Examples
    --------
    .. code:: python

        >>> from inspect import isclass
        >>> from nanoutils import import_factory

        >>> func = import_factory(isclass)
        >>> func('builtins.dict')
        <class 'dict'>

        >>> func('builtins.len')
        Traceback (most recent call last):
          ...
        RuntimeError: Passing <built-in function len> to isclass() failed to return True

    Parameters
    ----------
    validate : :data:`Callable[[T], bool]<typing.Callable>`
        A callable used for validating the passed object.
    module : :class:`str`
        The :attr:`~definition.__module__` of the to-be returned function.

    Returns
    -------
    :data:`Callable[[str], T]<typing.Callable>`
        A function for importing the passed object and validating it using **validate**.

    See Also
    --------
    :func:`nanoutils.get_importable`
        Import an object and, optionally, validate it using **validate**.

    """  # noqa: E501
    if not callable(validate):
        raise TypeError("'validate' expected a callable; observed type: "
                        f"{validate.__class__.__name__!r}")

    val_name = get_func_name(validate, repr_fallback=True)

    ret = PartialPrepend(get_importable, validate=validate)
    name_fallback = f'{ret.func.__class__.__name__}(...)'

    ret.__name__: str = getattr(ret.func, '__name__', name_fallback)  # type: ignore
    ret.__qualname__: str = getattr(ret.func, '__qualname__', name_fallback)  # type: ignore
    ret.__module__ = module
    ret.__doc__ = f'Return :code:`nanoutils.get_importable(string, validate={val_name})`.'
    ret.__signature__ = inspect.Signature(  # type: ignore
        parameters=[inspect.Parameter('string', _POK, annotation=str)],
        return_annotation=object
    )
    return ret


__doc__ = construct_api_doc(globals())



"""A private module for containing the :class:`nanoutils.SetAttr` class.

Notes
-----
:class:`~nanoutils.SetAttr` should be imported from
either :mod:`nanoutils` or :mod:`nanoutils.utils`.

"""

import reprlib
from types import TracebackType
from typing import Generic, TypeVar, NoReturn, Dict, Any, Optional, Type, List
from threading import RLock

__all__: List[str] = []

_T1 = TypeVar('_T1')
_T2 = TypeVar('_T2')
_ST = TypeVar('_ST', bound='SetAttr')


class SetAttr(Generic[_T1, _T2]):
    """A context manager for temporarily changing an attribute's value.

    The :class:`SetAttr` context manager is thread-safe, reusable and reentrant.

    Warnings
    --------
    Note that while :meth:`SetAttr.__enter__` and :meth:`SetAttr.__exit__` are thread-safe,
    the same does *not* hold for :meth:`SetAttr.__init__`.

    Examples
    --------
    .. code:: python

        >>> from nanoutils import SetAttr

        >>> class Test:
        ...     a = False

        >>> print(Test.a)
        False

        >>> set_attr = SetAttr(Test, 'a', True)
        >>> with set_attr:
        ...     print(Test.a)
        True

    """

    __slots__ = ('__weakref__', '_obj', '_name', '_value', '_value_old', '_lock', '_hash')

    @property
    def obj(self) -> _T1:
        """:class:`object`: Get the to-be modified object."""
        return self._obj

    @property
    def name(self) -> str:
        """:class:`str`: Get the name of the to-be modified attribute."""
        return self._name

    @property
    def value(self) -> _T2:
        """:class:`object`: Get the value to-be assigned to the :attr:`name` attribute of :attr:`SetAttr.obj`."""  # noqa: E501
        return self._value

    @property
    def attr(self) -> _T2:
        """:class:`object`: Get or set the :attr:`~SetAttr.name` attribute of :attr:`SetAttr.obj`."""  # noqa: E501
        return getattr(self.obj, self.name)  # type: ignore

    @attr.setter
    def attr(self, value: _T2) -> None:
        with self._lock:
            setattr(self.obj, self.name, value)

    def __init__(self, obj: _T1, name: str, value: _T2) -> None:
        """Initialize the :class:`SetAttr` context manager.

        Parameters
        ----------
        obj : :class:`object`
            The to-be modified object.
            See :attr:`SetAttr.obj`.
        name : :class:`str`
            The name of the to-be modified attribute.
            See :attr:`SetAttr.name`.
        value : :class:`object`
            The value to-be assigned to the **name** attribute of **obj**.
            See :attr:`SetAttr.value`.


        :rtype: :data:`None`

        """
        self._obj = obj
        self._name = name
        self._value = value

        self._value_old = self.attr
        self._lock = RLock()

    @reprlib.recursive_repr()
    def __repr__(self) -> str:
        """Implement :class:`str(self)<str>` and :func:`repr(self)<repr>`."""
        obj = object.__repr__(self.obj)
        value = reprlib.repr(self.value)
        return f'{self.__class__.__name__}(obj={obj}, name={self.name!r}, value={value})'

    def __eq__(self, value: object) -> bool:
        """Implement :meth:`self == value<object.__eq__>`."""
        if type(self) is not type(value):
            return False
        return self.obj is value.obj and self.name == value.name and self.value == value.value  # type: ignore  # noqa: E501

    def __reduce__(self) -> NoReturn:
        """A helper for :mod:`pickle`.

        Warnings
        --------
        Unsupported operation, raises a :exc:`TypeError`.

        """
        raise TypeError(f"can't pickle {self.__class__.__name__} objects")

    def __copy__(self: _ST) -> _ST:
        """Implement :func:`copy.copy(self)<copy.copy>`."""
        return self

    def __deepcopy__(self: _ST, memo: Optional[Dict[int, Any]] = None) -> _ST:
        """Implement :func:`copy.deepcopy(self, memo=memo)<copy.deepcopy>`."""
        return self

    def __hash__(self) -> int:
        """Implement :func:`hash(self)<hash>`.

        Warnings
        --------
        A :exc:`TypeError` will be raised if :attr:`SetAttr.value` is not hashable.

        """
        try:
            return self._hash
        except AttributeError:
            args = (type(self), (id(self.obj), self.name, self.value))
            self._hash: int = hash(args)
            return self._hash

    def __enter__(self) -> None:
        """Enter the context manager, modify :attr:`SetAttr.obj`."""
        self.attr = self.value

    def __exit__(self, exc_type: Optional[Type[BaseException]],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        """Exit the context manager, restore :attr:`SetAttr.obj`."""
        self.attr = self._value_old



"""Types related to the builtin :mod:`typing<python:typing>` module.

Contains aliases for ``python >= 3.8`` exclusive objects related to typing.

See Also
--------
.. image:: https://badge.fury.io/py/typing-extensions.svg
    :target: https://badge.fury.io/py/typing-extensions

The typing module: Support for gradual typing as defined by `PEP 484 <https://www.python.org/dev/peps/pep-0484/>`_.

At large scale, the structure of the module is following:

* Imports and exports, all public names should be explicitly added to :data:`__all__`.
* Internal helper functions: these should never be used in code outside this module.
* :class:`~typing._SpecialForm` and its instances (special forms): :data:`~typing.Any`,
  :data:`~typing.NoReturn`, :data:`~typing.ClassVar`, :data:`~typing.Union` and :data:`~typing.Optional`.
* Two classes whose instances can be type arguments in addition to types:
  :class:`~typing.ForwardRef` and :class:`~typing.TypeVar`.
* The core of internal generics API: _GenericAlias and _VariadicGenericAlias, the latter is
  currently only used by :data:`~typing.Tuple` and :data:`~typing.Callable`.
  All subscripted types like :code:`X[int]`, :code:`Union[int, str]`,
  *etc.*, are instances of either of these classes.
* The public counterpart of the generics API consists of two classes: :class:`~typing.Generic`
  and :class:`~typing.Protocol`.
* Public helper functions: :func:`~typing.get_type_hints`, :func:`~typing.overload`,
  :func:`~typing.cast`, :func:`~typing.no_type_check`, :func:`~typing.no_type_check_decorator`.
* Generic aliases for :mod:`collections.abc` ABCs and few additional protocols.
* Special types: :func:`~typing.NewType`, :class:`~typing.NamedTuple` and :class:`~typing.TypedDict`.
* Wrapper submodules for :mod:`re` and :mod:`io` related types.


Index
-----
========================================== ======================================================================================================================================
:data:`~typing.Literal`                    Special typing form to define literal types (a.k.a. value types).
:data:`~typing.Final`                      Special typing construct to indicate final names to type checkers.
:func:`~typing.final`                      A decorator to indicate final methods and final classes.
:class:`~typing.Protocol`                  Base class for protocol classes.
:class:`~typing.SupportsIndex`             An ABC with one abstract method :meth:`~object.__index__()`.
:class:`~typing.TypedDict`                 A simple typed name space. At runtime it is equivalent to a plain :class:`dict`.
:func:`~typing.runtime_checkable`          Mark a protocol class as a runtime protocol, so that it an be used with :func:`isinstance()` and :func:`issubclass()`.
:data:`~nanoutils.PathType`                An annotation for `path-like <https://docs.python.org/3/glossary.html#term-path-like-object>`_ objects.
:data:`~numpy.typing.ArrayLike`            Objects that can be converted to arrays (see :class:`numpy.ndarray`).
:data:`~numpy.typing.DTypeLike`            Objects that can be converted to dtypes (see :class:`numpy.dtype`).
:data:`ShapeLike<numpy.typing._ShapeLike>` Objects that can serve as valid array shapes.
========================================== ======================================================================================================================================

API
---
.. currentmodule:: nanoutils
.. data:: PathType
    :value: typing.Union[str, bytes, os.PathLike]

    An annotation for `path-like <https://docs.python.org/3/glossary.html#term-path-like-object>`_ objects.

"""  # noqa: E501

import os
import sys
from typing import Union, TYPE_CHECKING

if sys.version_info < (3, 8):
    from typing_extensions import Literal, Final, final, Protocol, TypedDict, runtime_checkable, SupportsIndex  # noqa: E501
else:
    from typing import Literal, Final, final, Protocol, TypedDict, SupportsIndex, runtime_checkable

if TYPE_CHECKING:
    # Requires numpy >= 1.20
    from numpy.typing import ArrayLike, DTypeLike, _ShapeLike as ShapeLike
    from numpy.typing import DTypeLike as DtypeLike

else:
    ArrayLike = 'numpy.typing.ArrayLike'
    DTypeLike = 'numpy.typing.DTypeLike'
    DtypeLike = DTypeLike
    ShapeLike = 'numpy.typing._ShapeLike'

__all__ = [
    'Literal', 'Final', 'final', 'Protocol', 'SupportsIndex', 'TypedDict', 'runtime_checkable',

    'PathType',

    'ArrayLike', 'DTypeLike', 'DtypeLike', 'ShapeLike'
]

# See https://github.com/python/typeshed/blob/master/stdlib/3/os/path.pyi.
PathType = Union[str, bytes, os.PathLike]



"""A module with empty (immutable) containers and iterables.

Can be used as default arguments for functions.

Index
-----
===================================== =================================================
:data:`~nanoutils.EMPTY_CONTAINER`     An empty :class:`~collections.abc.Container`.
:data:`~nanoutils.EMPTY_COLLECTION`    An empty :class:`~collections.abc.Collection`.
:data:`~nanoutils.EMPTY_SET`           An empty :class:`~collections.abc.Set`.
:data:`~nanoutils.EMPTY_SEQUENCE`      An empty :class:`~collections.abc.Sequence`.
:data:`~nanoutils.EMPTY_MAPPING`       An empty :class:`~collections.abc.Mapping`.
===================================== =================================================

API
---
.. currentmodule:: nanoutils
.. data:: EMPTY_CONTAINER
    :type: Container
    :value: frozenset()

    An immutable empty :class:`~collections.abc.Container`.

.. data:: EMPTY_COLLECTION
    :type: Collection
    :value: frozenset()

    An immutable empty :class:`~collections.abc.Collection`.

.. data:: EMPTY_SET
    :type: Set
    :value: frozenset()

    An immutable empty :class:`~collections.abc.Set`.

.. data:: EMPTY_SEQUENCE
    :type: Sequence
    :value: tuple()

    An immutable empty :class:`~collections.abc.Sequence`.

.. data:: EMPTY_MAPPING
    :type: Mapping
    :value: types.MappingProxyType({})

    An immutable empty :class:`~collections.abc.Mapping`.

"""

from types import MappingProxyType
from typing import Mapping, Collection, Sequence, AbstractSet, Container, Any

__all__ = [
    'EMPTY_SEQUENCE', 'EMPTY_MAPPING', 'EMPTY_COLLECTION', 'EMPTY_SET', 'EMPTY_CONTAINER'
]

EMPTY_SEQUENCE: Sequence[Any] = ()
EMPTY_MAPPING: Mapping[Any, Any] = MappingProxyType({})
EMPTY_COLLECTION: Collection[Any] = frozenset()
EMPTY_SET: AbstractSet[Any] = frozenset()
EMPTY_CONTAINER: Container[Any] = frozenset()



"""An abstract container for reading and writing files.

Index
-----
.. currentmodule:: nanoutils
.. autosummary::
    file_to_context
    AbstractFileContainer

API
---
.. autofunction:: file_to_context
.. autoclass:: AbstractFileContainer
    :members: read, _read, _read_postprocess, write, _write

"""

from __future__ import annotations

import sys
from abc import ABCMeta, abstractmethod
from codecs import decode, encode
from functools import partial
from contextlib import nullcontext
from typing import Dict, Any, AnyStr, Callable, ContextManager, TypeVar, overload, IO, Type

from .typing_utils import PathType, Literal, final
from .utils import _null_func

__all__ = ['AbstractFileContainer', 'file_to_context']

_ST = TypeVar('_ST', bound='AbstractFileContainer')

#: A Literal with accepted values for opening path-like objects in :class:`str` mode.
#: See https://github.com/python/typeshed/blob/master/stdlib/3/io.pyi.
_OpenTextMode = Literal[
    'r', 'r+', '+r', 'rt', 'tr', 'rt+', 'r+t', '+rt', 'tr+', 't+r', '+tr',
    'w', 'w+', '+w', 'wt', 'tw', 'wt+', 'w+t', '+wt', 'tw+', 't+w', '+tw',
    'a', 'a+', '+a', 'at', 'ta', 'at+', 'a+t', '+at', 'ta+', 't+a', '+ta',
    'x', 'x+', '+x', 'xt', 'tx', 'xt+', 'x+t', '+xt', 'tx+', 't+x', '+tx',
    'U', 'rU', 'Ur', 'rtU', 'rUt', 'Urt', 'trU', 'tUr', 'Utr',
]

#: A Literal with accepted values for opening path-like objects in :class:`bytes` mode.
#: See https://github.com/python/typeshed/blob/master/stdlib/3/io.pyi.
_OpenBinaryMode = Literal[
    'rb+', 'r+b', '+rb', 'br+', 'b+r', '+br',
    'wb+', 'w+b', '+wb', 'bw+', 'b+w', '+bw',
    'ab+', 'a+b', '+ab', 'ba+', 'b+a', '+ba',
    'xb+', 'x+b', '+xb', 'bx+', 'b+x', '+bx',
    'wb', 'bw',
    'ab', 'ba',
    'xb', 'bx',
    'rb', 'br',
    'rbU', 'rUb', 'Urb', 'brU', 'bUr', 'Ubr',
]


@overload
def file_to_context(
    file: IO[AnyStr],
    **kwargs: Any,
) -> ContextManager[IO[AnyStr]]:
    ...
@overload  # noqa: E302
def file_to_context(  # type: ignore[misc]
    file: int | PathType,
    *,
    mode: _OpenTextMode = ...,
    **kwargs: Any,
) -> ContextManager[IO[str]]:
    ...
@overload  # noqa: E302
def file_to_context(
    file: int | PathType,
    *,
    mode: _OpenBinaryMode = ...,
    **kwargs: Any,
) -> ContextManager[IO[bytes]]:
    ...
def file_to_context(file, **kwargs):  # noqa: E302
    r"""Take a path- or file-like object and return an appropiate context manager.

    Passing a path-like object will supply it to :func:`open`,
    while passing a file-like object will pass it to :func:`contextlib.nullcontext`.

    Examples
    --------
    .. code:: python

        >>> from io import StringIO
        >>> from nanoutils import file_to_context

        >>> path_like = 'file_name.txt'
        >>> file_like = StringIO('this is a file-like object')

        >>> context1 = file_to_context(file_like)
        >>> with context1 as f1:
        ...     ...  # doctest: +SKIP

        >>> context2 = file_to_context(path_like)  # doctest: +SKIP
        >>> with context2 as f2:  # doctest: +SKIP
        ...     ... # insert operations here  # doctest: +SKIP

    Parameters
    ----------
    file : :class:`str`, :class:`bytes`, :class:`os.PathLike` or :class:`~typing.IO`
        A `path- <https://docs.python.org/3/glossary.html#term-path-like-object>`_ or
        `file-like <https://docs.python.org/3/glossary.html#term-file-object>`_ object.
    **kwargs : :data:`~typing.Any`
        Further keyword arguments for :func:`open`.
        Only relevant if **file** is a path-like object.

    Returns
    -------
    :class:`ContextManager[IO]<typing.ContextManager>`
        An initialized context manager.
        Entering the context manager will return a file-like object.

    """
    # path-like object
    try:
        return open(file, **kwargs)

    # a file-like object (hopefully)
    except TypeError:
        return nullcontext(file)


class AbstractFileContainer(metaclass=ABCMeta):
    """An abstract container for reading and writing files.

    Two public methods are defined within this class:

    * :meth:`AbstractFileContainer.read`: Construct a new instance from this object's class by
      reading the content to a file or file object.
      How the content of the to-be read file is parsed has to be defined in the
      :meth:`AbstractFileContainer._read` abstract method.
      Additional post processing, after the new instance has been created, can be performed
      with :meth:`AbstractFileContainer._read_postprocess`
    * :meth:`AbstractFileContainer.write`: Write the content of this instance to an opened
      file or file object.
      How the content of the to-be exported class instance is parsed has to be defined in
      the :meth:`AbstractFileContainer._write`

    Examples
    --------
    .. code:: python

        >>> from io import StringIO
        >>> from nanoutils import AbstractFileContainer

        >>> class SubClass(AbstractFileContainer):
        ...     def __init__(self, value: str):
        ...         self.value = value
        ...
        ...     @classmethod
        ...     def _read(cls, file_obj, decoder):
        ...         value = decoder(file_obj.read())
        ...         return {'value': value}
        ...
        ...     def _write(self, file_obj, encoder):
        ...         value = encoder(self.value)
        ...         file_obj.write(value)

        >>> file1 = StringIO('This is a file-like object')
        >>> file2 = StringIO()

        >>> obj = SubClass.read(file1)
        >>> obj.write(file2)

        >>> print(file2.getvalue())
        This is a file-like object

    """

    __slots__ = ()

    @abstractmethod
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize an instance."""
        raise NotImplementedError("Trying to call an abstract method")

    @final
    @classmethod
    def read(
        cls: Type[_ST],
        file: PathType | IO[Any],
        bytes_decoding: None | str = None,
        **kwargs: Any,
    ) -> _ST:
        r"""Construct a new instance from this object's class by reading the content of **file**.

        Parameters
        ----------
        file : :class:`str`, :class:`bytes`, :class:`os.PathLike` or :class:`~typing.IO`
            A `path- <https://docs.python.org/3/glossary.html#term-path-like-object>`_ or
            `file-like <https://docs.python.org/3/glossary.html#term-file-object>`_ object.
        bytes_decoding : :class:`str`, optional
            The type of encoding to use when reading from **file**
            when it will be/is be opened in :class:`bytes` mode.
            This value should be left empty otherwise.
        \**kwargs : :data:`~typing.Any`
            Further keyword arguments for :func:`open`.
            Only relevant if **file** is a path-like object.

        Returns
        -------
        :class:`nanoutils.AbstractFileContainer`
            A new instance constructed from **file**.

        """  # noqa
        kwargs.setdefault('mode', 'r')
        context_manager = file_to_context(file, **kwargs)

        with context_manager as f:
            try:
                assert f.readable()
            except (TypeError, AttributeError) as ex:
                raise TypeError("'file' expected a file- or path-like object; "
                                f"observed type: {file.__class__.__name__!r}") from ex
            except AssertionError:
                f.read()  # This will raise an :exc:`io.UnsupportedOperation`

            if bytes_decoding is None:
                decoder: Callable[[Any], str] = _null_func
            else:
                decoder = partial(decode, encoding=bytes_decoding)  # type: ignore
            cls_dict = cls._read(f, decoder)

        ret = cls(**cls_dict)
        ret._read_postprocess()
        return ret

    @classmethod
    @abstractmethod
    def _read(cls, file_obj: IO[AnyStr], decoder: Callable[[AnyStr], str]) -> Dict[str, Any]:
        r"""A helper function for :meth:`~AbstractFileContainer.read`.

        Parameters
        ----------
        file_obj : :class:`IO[AnyStr]<typing.IO>`
            A file-like object opened in read mode.
        decoder : :data:`Callable[[AnyStr], str]<typing.Callable>`
            A function for converting the items of **file_obj** into strings.

        Returns
        -------
        :class:`Dict[str, Any]<typing.Dict>`
            A dictionary with keyword arguments for a new instance of this objects' class.

        See Also
        --------
        :meth:`AbstractFileContainer.read`
            Construct a new instance from this object's class by reading the content of **file**.

        """
        raise NotImplementedError('Trying to call an abstract method')

    def _read_postprocess(self) -> None:
        r"""Post process new instances created by :meth:`~AbstractFileContainer.read`.

        :rtype: :data:`None`

        See Also
        --------
        :meth:`AbstractFileContainer.read`
            Construct a new instance from this object's class by reading the content of **file**.

        """
        pass

    @final
    def write(
        self,
        file: PathType | IO[Any] = sys.stdout,
        bytes_encoding: None | str = None,
        **kwargs: Any,
    ) -> None:
        r"""Write the content of this instance to **file**.

        Parameters
        ----------
        file : :class:`str`, :class:`bytes`, :class:`os.PathLike` or :class:`~typing.IO`
            A `path- <https://docs.python.org/3/glossary.html#term-path-like-object>`_ or
            `file-like <https://docs.python.org/3/glossary.html#term-file-object>`_ object.
            Defaults to :data:`sys.stdout` if not specified.
        bytes_encoding : :class:`str`, optional
            The type of encoding to use when writing to **file**
            when it will be/is be opened in :class:`bytes` mode.
            This value should be left empty otherwise.
        \**kwargs : :data:`~typing.Any`
            Further keyword arguments for :func:`open`.
            Only relevant if **file** is a path-like object.


        :rtype: :data:`None`

        """
        kwargs.setdefault('mode', 'w')
        context_manager = file_to_context(file, **kwargs)

        with context_manager as f:
            try:
                assert f.writable()
            except (TypeError, AttributeError) as ex:
                raise TypeError("'file' expected a file- or path-like object; "
                                f"observed type: {file.__class__.__name__!r}") from ex
            except AssertionError:
                f.write(None)  # This will raise an :exc:`io.UnsupportedOperation`

            if bytes_encoding is None:
                encoder: Callable[[str], str | bytes] = _null_func
            else:
                encoder = partial(encode, encoding=bytes_encoding)
            self._write(f, encoder)

    @abstractmethod
    def _write(self, file_obj: IO[AnyStr], encoder: Callable[[str], AnyStr]) -> None:
        r"""A helper function for :meth:`~AbstractFileContainer.write`.

        Parameters
        ----------
        file_obj : :class:`IO[AnyStr]<typing.IO>`
            A file-like object opened in write mode.
        encoder : :class:`Callable[[str], AnyStr]<typing.Callable>`
            A function for converting strings into either :class:`str` or :class:`bytes`,
            the exact type matching that of **file_obj**.


        :rtype: :data:`None`

        See Also
        --------
        :meth:`AbstractFileContainer.write`:
            Write the content of this instance to **file**.

        """
        raise NotImplementedError('Trying to call an abstract method')



"""The **Nano-Utils** package."""

# flake8: noqa: F403,F401

from .__version__ import __version__

from . import (typing_utils, empty, utils, testing_utils, numpy_utils, schema,
               file_container, hdf5_utils, yaml_utils)
from .typing_utils import *
from .empty import *
from .utils import *
from .testing_utils import *
from .numpy_utils import *
from .file_container import *
from .schema import (
    Default, Formatter, supports_float, supports_int,
    isinstance_factory, issubclass_factory, import_factory
)
from .hdf5_utils import *
from .yaml_utils import *

__author__ = 'B. F. van Beek'
__email__ = 'b.f.van.beek@vu.nl'
version_info = VersionInfo.from_str(__version__)

__all__ = []
__all__ += typing_utils.__all__
__all__ += empty.__all__
__all__ += utils.__all__
__all__ += testing_utils.__all__
__all__ += numpy_utils.__all__
__all__ += schema.__all__
__all__ += file_container.__all__
__all__ += hdf5_utils.__all__
__all__ += yaml_utils.__all__
__all__.sort()



"""A module for the :class:`DTypeMapping` and :class:`MutableDTypeMapping` classes."""

from __future__ import annotations

import abc
import sys
import copy
import textwrap
import reprlib
from collections.abc import Iterator, KeysView, ValuesView, ItemsView, Callable
from typing import TypeVar, Union, Any, overload, NoReturn, TYPE_CHECKING

if sys.version_info >= (3, 8):
    import functools
    from pprint import pformat as _pformat
    pformat = functools.partial(_pformat, sort_dicts=False)
else:
    from pprint import pformat

if sys.version_info >= (3, 9):
    from collections.abc import Mapping, MutableMapping, Iterable
    from builtins import tuple as Tuple
else:
    from typing import Mapping, MutableMapping, Iterable, Tuple

from .utils import positional_only
from .typing_utils import Protocol, runtime_checkable

__all__ = ["UserMapping", "MutableUserMapping", "_DictLike", "_SupportsKeysAndGetItem"]

_SENTINEL = object()

_T = TypeVar("_T")
_ST1 = TypeVar("_ST1", bound="UserMapping[Any, Any]")
_ST2 = TypeVar("_ST2", bound="MutableUserMapping[Any, Any]")
_KT = TypeVar("_KT")
_VT = TypeVar("_VT")
_VT_co = TypeVar("_VT_co", covariant=True)


@runtime_checkable
class _SupportsKeysAndGetItem(Protocol[_KT, _VT_co]):
    """A protocol for objects supporting :meth:`~dict.keys` and `~object.__getitem__`."""

    if not TYPE_CHECKING:
        __slots__: str | Iterable[str] = ()

    @abc.abstractmethod
    def keys(self) -> Iterable[_KT]:
        """Return an iterable of keys."""
        raise NotImplementedError

    @abc.abstractmethod
    def __getitem__(self, __key: _KT) -> _VT_co:
        """Implement :meth:`self[key] <object.__getitem__>`."""
        raise NotImplementedError


_DictLike = Union[
    _SupportsKeysAndGetItem[_KT, _VT_co],
    Iterable[Tuple[_KT, _VT_co]],
]


class UserMapping(Mapping[_KT, _VT_co]):
    """Base class for user-defined immutable mappings."""

    __slots__: str | Iterable[str] = ("__weakref__", "_dict", "_hash")

    _dict: dict[_KT, _VT_co]
    _hash: int

    @positional_only
    def __init__(
        self,
        __iterable: None | _DictLike[_KT, _VT_co] = None,
        **kwargs: _VT_co,
    ) -> None:
        """Initialize the instance."""
        if __iterable is None:
            self._dict = kwargs  # type: ignore[assignment]
        else:
            self._dict = dict(__iterable, **kwargs)

    @classmethod
    def _reconstruct(cls: type[_ST1], dct: dict[_KT, _VT_co]) -> _ST1:
        """Alternative constructor without argument validation."""
        self = cls.__new__(cls)
        self._dict = dct
        return self  # type: ignore[no-any-return]

    def __reduce__(self: _ST1) -> tuple[
        Callable[[dict[_KT, _VT_co]], _ST1],
        tuple[dict[_KT, _VT_co]],
    ]:
        """Helper for :mod:`pickle`."""
        cls = type(self)
        return cls._reconstruct, (self._dict,)

    def copy(self: _ST1) -> _ST1:
        """Return a deep copy of this instance."""
        return copy.deepcopy(self)

    @reprlib.recursive_repr(fillvalue='...')
    def __repr__(self) -> str:
        """Implement :func:`repr(self) <repr>`."""
        cls = type(self)
        width = 80 - 2 - len(cls.__name__)
        dct_repr = pformat(self._dict, width=width)
        if len(dct_repr) <= width:
            return f"{cls.__name__}({dct_repr})"
        else:
            dct_repr2 = textwrap.indent(dct_repr[1:-1], 3 * " ")
            return f"{cls.__name__}({{\n {dct_repr2},\n}})"

    def __hash__(self) -> int:
        """Implement :func:`hash(self) <hash>`.

        Raises
        ------
        :exc:`TypeError` : raised when not all values are hashable.

        """
        try:
            return self._hash
        except AttributeError:
            pass
        self._hash = hash(frozenset(self.items()))
        return self._hash

    def __eq__(self, other: object) -> bool:
        """Implement :meth:`self == other <object.__eq__>`."""
        if isinstance(other, UserMapping):
            return self._dict == other._dict
        elif isinstance(other, Mapping):
            return self._dict == other
        else:
            return NotImplemented

    def __getitem__(self, key: _KT) -> _VT_co:
        """Implement :meth:`self[key] <object.__getitem__>`."""
        return self._dict[key]

    def __iter__(self) -> Iterator[_KT]:
        """Implement :func:`iter(self) <iter>`."""
        return iter(self._dict)

    def __len__(self) -> int:
        """Implement :func:`len(self) <len>`."""
        return len(self._dict)

    def __contains__(self, key: object) -> bool:
        """Implement :meth:`key in self <object.__contains__>`."""
        return key in self._dict

    def keys(self) -> KeysView[_KT]:
        """Return a set-like object containing all keys."""
        return self._dict.keys()

    def items(self) -> ItemsView[_KT, _VT_co]:
        """Return a set-like object containing all key/value pairs."""
        return self._dict.items()

    def values(self) -> ValuesView[_VT_co]:
        """Return a collection containing all values."""
        return self._dict.values()

    @overload
    def get(self, key: _KT, default: None = ...) -> _VT_co | None: ...
    @overload
    def get(self, key: _KT, default: _T) -> _VT_co | _T: ...

    def get(self, key, default=None):
        """Return the value for key if the key is present, else default."""
        return self._dict.get(key, default)

    @overload
    @classmethod
    def fromkeys(cls, iterable: Iterable[_T]) -> UserMapping[_T, None]: ...
    @overload
    @classmethod
    def fromkeys(cls, iterable: Iterable[_T], value: _VT) -> UserMapping[_T, _VT]: ...

    @classmethod
    def fromkeys(cls, iterable, value=None):
        """Create a new dictionary with keys from iterable and values set to default."""
        dct = dict.fromkeys(iterable, value)
        return cls._reconstruct(dct)

    if sys.version_info >= (3, 8):
        def __reversed__(self) -> Iterator[_KT]:
            """Implement :func:`reversed(self) <reversed>`."""
            return reversed(self._dict)

    if sys.version_info >= (3, 9):
        def __or__(self: _ST1, other: Mapping[_KT, _VT_co]) -> _ST1:
            """Implement :meth:`self | other <object.__or__>`."""
            cls = type(self)
            if not isinstance(other, Mapping):
                return NotImplemented
            elif isinstance(other, UserMapping):
                return cls._reconstruct(self._dict | other._dict)
            else:
                return cls._reconstruct(self._dict | other)

        def __ror__(self: _ST1, other: Mapping[_KT, _VT_co]) -> _ST1:
            """Implement :meth:`other | self <object.__ror__>`."""
            cls = type(self)
            if not isinstance(other, Mapping):
                return NotImplemented
            elif isinstance(other, UserMapping):
                return cls._reconstruct(other._dict | self._dict)
            else:
                return cls._reconstruct(other | self._dict)

        if not TYPE_CHECKING:
            def __ior__(self, other: Mapping[_KT, _VT_co]) -> NoReturn:
                """Implement :meth:`self |= other <object.__ior__>`.

                Warning
                -------
                Unsupported operation.

                """
                cls = type(self)
                raise TypeError(f"'|=' is not supported by {cls.__name__!r}; use '|' instead")


class MutableUserMapping(UserMapping[_KT, _VT], MutableMapping[_KT, _VT]):
    """Base class for user-defined mutable mappings."""

    __slots__: str | Iterable[str] = ()
    __hash__ = None  # type: ignore[assignment]

    def __copy__(self: _ST2) -> _ST2:
        """Implement :func:`copy.copy(self) <copy.copy>`."""
        return copy.deepcopy(self)

    def __setitem__(self, key: _KT, value: _VT) -> None:
        """Implement :meth:`self[key] = value <object.__setitem__>`."""
        self._dict[key] = value

    def __delitem__(self, key: _KT) -> None:
        """Implement :meth:`del self[key] <object.__delitem__>`."""
        del self._dict[key]

    def clear(self) -> None:
        """Remove all items from the mapping."""
        return self._dict.clear()

    def popitem(self) -> tuple[_KT, _VT]:
        """Remove and return a (key, value) pair as a 2-tuple.

        Pairs are returned in LIFO (last-in, first-out) order.
        Raises a :exc:`KeyError` if the mapping is empty.
        """
        return self._dict.popitem()

    @overload
    def pop(self, key: _KT) -> _VT: ...
    @overload
    def pop(self, key: _KT, default: _T = ...) -> _VT | _T: ...

    def pop(self, key, default=_SENTINEL):
        """Remove the specified key and return the corresponding value.

        If the key is not found, default is returned if given,
        otherwise a :exc:`KeyError` is raised.
        """
        if default is _SENTINEL:
            return self._dict.pop(key)
        else:
            return self._dict.pop(key, default)

    @positional_only
    def update(self, __iterable: None | _DictLike[_KT, _VT] = None, **kwargs: _VT) -> None:
        """Update the mapping from the passed mapping or iterable."""
        if __iterable is None:
            self._dict.update(**kwargs)
        else:
            self._dict.update(__iterable, **kwargs)  # type: ignore[arg-type]

    if sys.version_info >= (3, 9):
        def __ior__(self: _ST2, other: Mapping[_KT, _VT]) -> _ST2:
            """Implement :meth:`self |= other <object.__ior__>`."""
            if not isinstance(other, Mapping):
                return NotImplemented
            elif isinstance(other, UserMapping):
                self._dict |= other._dict
            else:
                self._dict |= other
            return self

    if TYPE_CHECKING:
        @overload  # type: ignore
        @classmethod
        def fromkeys(cls, iterable: Iterable[_T]) -> MutableUserMapping[_T, None | Any]: ...
        @overload
        @classmethod
        def fromkeys(cls, iterable: Iterable[_T], value: _VT) -> MutableUserMapping[_T, _VT]: ...



"""A :class:`yaml.Loader` subclass that dissallows for duplicate keys.

Index
-----
.. currentmodule:: nanoutils
.. autosummary::
    UniqueLoader

API
---
.. autoclass:: UniqueLoader

"""

from __future__ import annotations

from collections.abc import Hashable
from typing import Dict, Any, IO

from .utils import raise_if

try:
    import yaml
    try:
        from yaml import CLoader as Loader
    except ImportError:
        from yaml import Loader  # type: ignore[misc]
    YAML_EX: None | Exception = None
except Exception as ex:
    from builtins import object as Loader  # type: ignore[misc]
    YAML_EX = ex

__all__ = ['UniqueLoader']


class UniqueLoader(Loader):
    '''A :class:`yaml.Loader` subclass that dissallows for duplicate keys.

    Examples
    --------
    .. doctest:: python
        :skipif: YAML_EX is not None

        >>> import yaml
        >>> from nanoutils import UniqueLoader

        >>> STR = """
        ... a: 0
        ... a: 1
        ... """

        >>> yaml.load(STR, Loader=yaml.SafeLoader)
        {'a': 1}

        >>> yaml.load(STR, Loader=UniqueLoader)  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
          ...
        yaml.constructor.ConstructorError: while constructing a mapping
          in "<unicode string>", line 2, column 1
        found a duplicate key
          in "<unicode string>", line 3, column 1

    '''

    @raise_if(YAML_EX)
    def __init__(cls, stream: str | bytes | IO[str] | IO[bytes]) -> None:
        super().__init__(stream)

    def construct_mapping(self, node: yaml.MappingNode, deep: bool = False) -> Dict[Any, Any]:
        """Construct Convert the passed **node** into a :class:`dict`."""
        if not isinstance(node, yaml.MappingNode):
            raise yaml.constructor.ConstructorError(
                None, None, f"expected a mapping node, but found {node.id}", node.start_mark
            )

        mapping = {}
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            if not isinstance(key, Hashable):
                raise yaml.constructor.ConstructorError(
                    "while constructing a mapping", node.start_mark,
                    "found unhashable key", key_node.start_mark,
                )
            elif key in mapping:
                raise yaml.constructor.ConstructorError(
                    "while constructing a mapping", node.start_mark,
                    "found a duplicate key", key_node.start_mark,
                )

            value = self.construct_object(value_node, deep=deep)
            mapping[key] = value
        return mapping



"""Utility functions related to unit-testing.

Index
-----
.. currentmodule:: nanoutils
.. autosummary::
{autosummary}

API
---
{autofunction}

"""

from __future__ import annotations

import os
import shutil
import warnings
from typing import TypeVar, Callable, Any, AnyStr, Tuple, cast
from os.path import isdir, isfile, join
from functools import wraps

from .typing_utils import PathType
from .utils import construct_api_doc

__all__ = ['FileNotFoundWarning', 'delete_finally']

_FT = TypeVar('_FT', bound=Callable[..., Any])


class FileNotFoundWarning(ResourceWarning):
    """A :exc:`ResourceWarning` subclass for when a file or directory is requested but doesn’t exist."""  # noqa: E501


def _delete_finally(path: PathType, warn: bool = True) -> None:
    """Helper function for :func:`delete_finally`."""
    try:
        if isdir(path):
            shutil.rmtree(path)
        elif isfile(path):
            os.remove(path)
        elif warn:
            _warning = FileNotFoundWarning(f'No such file or directory: {path!r}')
            warnings.warn(_warning, stacklevel=3)

    # In case an unexpected exception is encountered
    except Exception as ex:
        _warning2 = RuntimeWarning(str(ex))
        _warning2.__cause__ = ex
        warnings.warn(_warning2, stacklevel=3)


def delete_finally(
    *paths: AnyStr | os.PathLike[AnyStr],
    prefix: None | AnyStr = None,
    warn: bool = True,
) -> Callable[[_FT], _FT]:
    r"""A decorater which deletes the specified files and/or directories after calling the deocrated function.

    Examples
    --------
    .. code:: python

        >>> import os
        >>> from nanoutils import delete_finally

        >>> file1 = 'file1.txt'
        >>> dir1 = 'dir1/'
        >>> os.path.isfile(file1) and os.path.isdir(dir1)  # doctest: +SKIP
        True

        >>> @delete_finally(file1, dir1)
        ... def func():
        ...     pass

        >>> func()  # doctest: +SKIP
        >>> os.path.isfile(file1) or os.path.isdir(dir1)  # doctest: +SKIP
        False

    Parameters
    ----------
    \*paths : :class:`str`, :class:`bytes` or :class:`os.PathLike`
        Path-like objects with the names of to-be deleted files and/or directories.
    prefix : :class:`str`, :class:`bytes` or :class:`os.PathLike`, optional
        The directory where all user specified **paths** are located.
        If :data:`None`, asume the files/directories are either absolute or
        located in the current working directory.
    warn : :class:`bool`
        If :data:`True` issue a :exc:`~nanoutils.FileNotFoundWarning` if
        a to-be deleted file or directory cannot be found.

    """  # noqa: E501
    if prefix is not None:
        _PATH_TUP: Tuple[PathType, ...] = tuple(join(prefix, path) for path in paths)
    else:
        _PATH_TUP = paths

    def decorator(func: _FT) -> _FT:
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            finally:
                for path in _PATH_TUP:
                    _delete_finally(path, warn)

        return cast(_FT, wrapper)
    return decorator


__doc__ = construct_api_doc(globals(), decorators={'delete_finally'})



"""A private module for containg the :class:`nanoutils.CatchErrors` class.

Notes
-----
:class:`~nanoutils.CatchErrors` should be imported from
either :mod:`nanoutils` or :mod:`nanoutils.utils`.

"""

from __future__ import annotations

import types
from typing import (
    TypeVar,
    Generic,
    Tuple,
    Type,
    List,
    cast,
    Iterable,
    Any,
    Callable,
)

from ._seq_view import SequenceView

_ET = TypeVar("_ET", bound=BaseException)
_ET2 = TypeVar("_ET2", bound=BaseException)

__all__ = ["CatchErrors"]


class CatchErrors(Generic[_ET]):
    """A re-usable context manager for catching and storing all exceptions of a given type.

    Examples
    --------
    .. code-block:: python

        >>> from nanoutils import CatchErrors

        >>> context = CatchErrors(AssertionError)

        >>> for i in range(3):
        ...     with context as exc_view:
        ...         assert False, i  # doctest: +SKIP
        ...         print(exc_view)
        SequenceView([AssertionError(0)])
        SequenceView([AssertionError(0), AssertionError(1)])
        SequenceView([AssertionError(0), AssertionError(1), AssertionError(2)])

    See Also
    --------
    :class:`contextlib.suppress`
        Context manager to suppress specified exceptions.

    """  # noqa: E501

    __slots__ = ("__weakref__", "_exceptions", "_caught_exceptions")

    @property
    def exceptions(self) -> Tuple[Type[_ET], ...]:
        """Get the to-be caught exception types."""
        return self._exceptions

    @property
    def caught_exceptions(self) -> SequenceView[_ET]:
        """Get a read-only view of all caught exceptions."""
        return SequenceView(self._caught_exceptions)

    def __init__(self, *exceptions: Type[_ET]) -> None:
        r"""Initialize this instance.

        Parameters
        ----------
        \*exceptions : :class:`type[BaseException] <type>`
            The to-be caught exceptions.

        """
        self._exceptions = exceptions
        self._caught_exceptions: List[_ET] = []

    def clear(self) -> None:
        """Clear all caught exceptions."""
        self._caught_exceptions = []

    def __repr__(self) -> str:
        """Implement :func:`repr(self) <repr>`."""
        name = type(self).__name__
        args = ", ".join(i.__name__ for i in self.exceptions)
        return f"{name}({args})"

    @classmethod
    def _reconstruct(
        cls: Type["CatchErrors[Any]"],
        exc_types: Iterable[Type[_ET2]],
        exc_list: List[_ET2],
    ) -> "CatchErrors[_ET2]":
        """Construct a new :class:`CatchErrors` instance; see :meth:`__reduce__`."""
        self: CatchErrors[_ET2] = cls(*exc_types)
        self._caught_exceptions = exc_list
        return self

    def __reduce__(self) -> Tuple[
        Callable[[Iterable[Type[_ET2]], List[_ET2]], "CatchErrors[_ET2]"],
        Tuple[Tuple[Type[_ET], ...], List[_ET]],
    ]:
        """Helper method for :mod:`pickle`."""
        cls = type(self)
        return cls._reconstruct, (self.exceptions, self._caught_exceptions)

    def __enter__(self) -> SequenceView[_ET]:
        """Enter the context manager.

        Returns
        -------
        :class:`nanoutils.SequenceView`
            A read-only sequence view with all caught exceptions.

        """
        return self.caught_exceptions

    def __exit__(
        self,
        exctype: None | Type[BaseException] = None,
        excinst: None | BaseException = None,
        exctb: None | types.TracebackType = None,
    ) -> bool:
        """Exit the context manager.

        Update :attr:`caught_exceptions` if an appropriate exception is encountered.
        """
        if exctype is not None and issubclass(exctype, self.exceptions):
            self._caught_exceptions.append(cast(_ET, excinst))
            return True
        return False



"""A private module for containg the :class:`LazyImporter` and :class:`MutableLazyImporter` classes.

Notes
-----
:class:`~nanoutils.LazyImporter` and :class:`~nanoutils.MutableLazyImporter`
should be imported from either :mod:`nanoutils` or :mod:`nanoutils.utils`.

"""  # noqa: E501

from __future__ import annotations

import sys
import types
import reprlib
import importlib
from collections.abc import Mapping, Callable, Iterable
from typing import Any, TypeVar, Type, Generic, cast, TYPE_CHECKING

_T = TypeVar("_T")

if TYPE_CHECKING:
    from typing_extensions import Protocol
    from typing import (
        overload,
        Collection,
        Union,
        Tuple,
        Dict,
        KeysView,
        ValuesView,
        ItemsView,
        Iterator,
    )

    _KT = TypeVar("_KT")
    _VT = TypeVar("_VT")
    _VT_co = TypeVar("_VT_co", covariant=True)
    _ST1 = TypeVar("_ST1", bound="LazyImporter[Any]")
    _ST2 = TypeVar("_ST2", bound="MutableLazyImporter[Any]")

    class _SupportsKeysAndGetItem(Protocol[_KT, _VT_co]):
        def __getitem__(self, __key: _KT) -> _VT_co: ...
        def keys(self) -> Iterable[_KT]: ...

    # A `Protocol` version of `types.MappingProxyType`
    class _MappingProtocol(Protocol[_KT, _VT], Collection[_KT]):
        __hash__: None  # type: ignore
        def __getitem__(self, k: _KT) -> _VT: ...
        def __contains__(self, o: object) -> bool: ...
        @overload
        def get(self, key: _KT) -> None | _VT: ...
        @overload
        def get(self, key: _KT, default: _T) -> _T | _VT: ...
        def items(self) -> ItemsView[_KT, _VT]: ...
        def keys(self) -> KeysView[_KT]: ...
        def values(self) -> ValuesView[_VT]: ...
        def copy(self) -> dict[_KT, _VT]: ...

        # TODO: Update `__or__` once mypy > 0.910 hits
        if sys.version_info >= (3, 9):
            def __reversed__(self) -> Iterator[_KT]: ...
            def __or__(self, __value: Mapping[_KT, _VT]) -> dict[_KT, _VT]: ...
            def __ror__(self, __value: Mapping[_KT, _VT]) -> dict[_KT, _VT]: ...

    _DictLike = Union[_SupportsKeysAndGetItem[str, str], Iterable[Tuple[str, str]]]
    _ReduceTuple = Tuple[Callable[[str, Dict[str, str]], _T], Tuple[str, Dict[str, str]]]

__all__ = ["MutableLazyImporter", "LazyImporter"]


class _CustomRepr(reprlib.Repr):
    def __init__(self) -> None:
        super().__init__()
        self.maxdict = 2

    def repr_mappingproxy(self, x, level):
        return self.repr_dict(x, level)


_repr = _CustomRepr().repr


class LazyImporter(Generic[_T]):
    """A class for lazilly importing objects.

    Parameters
    ----------
    module : :class:`types.ModuleType`
        The to-be wrapped module.
    imports : :class:`Mapping[str, str]<collections.abc.Mapping>`
        A mapping that maps the names of to-be lazzily imported objects
        to the names of their modules.

    Examples
    --------
    .. code-block:: python

        >>> from nanoutils import LazyImporter

        >>> __getattr__ = LazyImporter.from_name("nanoutils", {"Any": "typing"})
        >>> print(__getattr__)
        LazyImporter(module=nanoutils, imports={'Any': 'typing'})

        >>> __getattr__("Any")
        typing.Any

    """

    __slots__ = ("__weakref__", "_imports", "_module", "_hash")

    @property
    def module(self) -> types.ModuleType:
        """:class:`types.ModuleType`: Get the wrapped module."""
        return self._module

    @property
    def imports(self) -> _MappingProtocol[str, str]:
        """:class:`types.MappingProxyType[str, str]<types.MappingProxyType>`: Get a mapping that maps object names to their module name."""  # noqa: E501
        return self._imports

    def __init__(
        self: LazyImporter[Any],
        module: types.ModuleType,
        imports: _DictLike,
    ) -> None:
        """Initialize a new instance."""
        if not isinstance(module, types.ModuleType):
            raise TypeError(f"Expected a module, not {type(module).__name__}")
        self._module = module
        self._imports = cast("_MappingProtocol[str, str]", types.MappingProxyType(dict(imports)))

    @classmethod
    def from_name(cls: Type[_ST1], name: str, imports: _DictLike) -> _ST1:
        """Construct a new instance from the module **name**.

        Parameters
        ----------
        name : :class:`str`
            The name of the to-be wrapped module.
        imports : :class:`Mapping[str, str]<collections.abc.Mapping>`
            A mapping that maps the names of to-be lazzily imported objects
            to the names of their modules.

        Returns
        -------
        :class:`nanoutils.LazyImporter`
            A new LazyImporter instance or a subclass thereof.

        """
        if not isinstance(name, str):
            raise TypeError(f"Expected a string, not {type(name).__name__}")

        module = sys.modules.get(name)  # fast-path
        if module is None:
            module = importlib.import_module(name)
        return cls(module, imports)

    def __reduce__(self: _ST1) -> _ReduceTuple[_ST1]:
        """Helper for :mod:`pickle`."""
        cls = type(self)
        args = (self.module.__name__, self.imports.copy())
        return cls.from_name, args

    def __copy__(self: _ST1) -> _ST1:
        """Implement :func:`copy.copy(self)<copy.copy>`."""
        return self

    def __deepcopy__(self: _ST1, memo: None | dict[int, Any] = None) -> _ST1:
        """Implement :func:`copy.deepcopy(self, memo=memo)<copy.deepcopy>`."""
        return self

    def __hash__(self) -> int:
        """Implement :func:`hash(self)<hash>`."""
        try:
            return self._hash
        except AttributeError:
            self._hash: int = hash(self._module) ^ hash(frozenset(self.imports.items()))
            return self._hash

    def __eq__(self, value: object) -> bool:
        """Implement :meth:`self == value<object.__eq__>`."""
        cls = type(self)
        if not isinstance(value, cls):
            return NotImplemented
        return self.module is value.module and self.imports == value.imports

    def __repr__(self) -> str:
        """Implement :func:`repr(self)<repr>`."""
        name = type(self).__name__
        return f"{name}(module={self.module.__name__}, imports={_repr(self.imports)})"

    def __call__(self, name: str) -> _T:
        """Implement :func:`getattr(self.module, name)<getattr>`."""
        # Get the module-name associated with `name`
        try:
            module_name = self.imports[name]
        except KeyError:
            raise AttributeError(
                f"module {self.module.__name__!r} has no attribute {name!r}"
            ) from None

        # Import the module
        module = sys.modules.get(module_name)
        if module is None:
            module = importlib.import_module(module_name)

        # Extract `name` from `module`
        ret: _T = getattr(module, name)
        setattr(self.module, name, ret)
        return ret


class MutableLazyImporter(LazyImporter[_T]):
    """A subclass of :class:`~nanoutils.LazyImporter` with mutable :attr:`imports`.

    Parameters
    ----------
    module : :class:`types.ModuleType`
        The to-be wrapped module.
    imports : :class:`Mapping[str, str]<collections.abc.Mapping>`
        A mapping that maps the names of to-be lazzily imported objects
        to the names of their modules.

    Examples
    --------
    .. code-block:: python

        >>> from nanoutils import MutableLazyImporter

        >>> __getattr__ = MutableLazyImporter.from_name("nanoutils", {"Any": "typing"})
        >>> print(__getattr__)
        MutableLazyImporter(module=nanoutils, imports={'Any': 'typing'})

        >>> __getattr__("Any")
        typing.Any

        >>> del __getattr__.imports["Any"]
        >>> print(__getattr__)
        MutableLazyImporter(module=nanoutils, imports={})

        >>> __getattr__.imports = {"Hashable": "collections.abc"}
        >>> __getattr__("Hashable")
        <class 'collections.abc.Hashable'>

    """

    __hash__ = None  # type: ignore[assignment]

    @property
    def imports(self) -> dict[str, str]:
        """:class:`dict[str, str]<dict>`: Get or set the dictionary that maps object names to their module name.

        Setting a value will assign it as a copy.

        """  # noqa: E501
        return self._imports

    @imports.setter
    def imports(self, value: _DictLike) -> None:
        self._imports = dict(value)

    def __init__(
        self: MutableLazyImporter[Any],
        module: types.ModuleType,
        imports: _DictLike,
    ) -> None:
        """Initialize a :class:`MutableLazyImporter` instance."""
        if not isinstance(module, types.ModuleType):
            raise TypeError(f"Expected a module, not {module.__class__.__name__}")
        self._module = module
        self._imports: dict[str, str] = dict(imports)

    def __copy__(self: _ST2) -> _ST2:
        """Implement :func:`copy.copy(self)<copy.copy>`."""
        cls = type(self)
        ret: _ST2 = object.__new__(cls)
        ret._module = self.module
        ret._imports = self.imports
        return ret

    def __deepcopy__(self: _ST2, memo: None | dict[int, Any] = None) -> _ST2:
        """Implement :func:`copy.deepcopy(self, memo=memo)<copy.deepcopy>`."""
        cls = type(self)
        ret: _ST2 = object.__new__(cls)
        ret._module = self.module
        ret._imports = self.imports.copy()
        return ret

    def __reduce__(self: _ST2) -> _ReduceTuple[_ST2]:
        """Helper for :mod:`pickle`."""
        cls = type(self)
        return cls.from_name, (self.module.__name__, self.imports)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from typing import Dict

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

# To update the package version number, edit nanoutils/__version__.py
version: Dict[str, str] = {}
version_path = os.path.join(here, 'nanoutils', '__version__.py')
with open(version_path, encoding='utf-8') as f:
    exec(f.read(), version)

with open('README.rst', encoding='utf-8') as f:
    readme = f.read()

# Requirements for building the documentation
docs_require = [
    'sphinx>=2.4',
    'sphinx_rtd_theme',
]

# Requirements for building wheels
build_requires = [
    'twine',
    'wheel'
]

# Requirements for running tests
tests_no_optional_require = [
    'assertionlib>=3.2.1',
    'mypy>=0.900',
    'pytest>=5.4.0',
    'pytest-cov',
    'pytest-mypy>=0.6.2',
    'types-PyYAML',
]
tests_require = tests_no_optional_require + [
    'schema',
    'pyyaml',
    'h5py',
    'numpy',
]
tests_require += docs_require
tests_require += build_requires

setup(
    name='Nano-Utils',
    version=version['__version__'],
    description='Utility functions used throughout the various nlesc-nano repositories.',
    long_description=f'{readme}\n\n',
    long_description_content_type='text/x-rst',
    author=['Bas van Beek'],
    author_email='b.f.van.beek@vu.nl',
    url='https://github.com/nlesc-nano/Nano-Utils',
    packages=['nanoutils'],
    package_dir={'nanoutils': 'nanoutils'},
    package_data={'nanoutils': ['py.typed', '*.pyi']},
    include_package_data=True,
    license='Apache Software License',
    zip_safe=False,
    keywords=[
        'python-3',
        'python-3-7',
        'python-3-8',
        'python-3-9',
        'libraries'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development :: Libraries',
        'Typing :: Typed'
    ],
    python_requires='>=3.7',
    install_requires=[
        'typing_extensions>=3.10.0; python_version<"3.8"',
    ],
    setup_requires=['pytest-runner'] + docs_require,
    tests_require=tests_require,
    extras_require={
        'doc': docs_require,
        'test': tests_require,
        'test_no_optional': tests_no_optional_require,
        'build': build_requires,
    }
)



"""General utility functions.

Index
-----
.. currentmodule:: nanoutils
.. autosummary::
{autosummary}

API
---
{autofunction}

"""

from __future__ import annotations

import re
import warnings
import importlib
import inspect
from types import ModuleType
from functools import wraps
from typing import (
    List,
    Any,
    Tuple,
    Callable,
    TypeVar,
    Iterable,
    Dict,
    Container,
    Mapping,
    NamedTuple,
    MutableMapping,
    Collection,
    cast,
    overload
)

from .empty import EMPTY_CONTAINER

__all__ = [
    'PartialPrepend',
    'SetAttr',
    'VersionInfo',
    'group_by_values',
    'get_importable',
    'construct_api_doc',
    'split_dict',
    'get_func_name',
    'set_docstring',
    'raise_if',
    'ignore_if',
    'SequenceView',
    'CatchErrors',
    'LazyImporter',
    'MutableLazyImporter',
    'positional_only',
    'UserMapping',
    'MutableUserMapping',
]

_T = TypeVar('_T')
_KT = TypeVar('_KT')
_VT = TypeVar('_VT')
_FT = TypeVar('_FT', bound=Callable[..., Any])


def get_func_name(
    func: Callable[..., Any],
    prepend_module: bool = False,
    repr_fallback: bool = False,
) -> str:
    """Extract and return the name of **func**.

    A total of three attempts are performed at retrieving the passed functions name:

    1. Return the functions qualified name (:attr:`~definition.__qualname__`).
    2. Return the functions name (:attr:`~definition.__name__`).
    3. Return the (called) name of the functions type.

    Examples
    --------
    .. code:: python

        >>> from functools import partial
        >>> from nanoutils import get_func_name

        >>> def func1():
        ...     pass

        >>> class Class():
        ...     def func2(self):
        ...         pass

        >>> func3 = partial(len)

        >>> get_func_name(func1)
        'func1'

        >>> get_func_name(func1, prepend_module=True)  # doctest: +SKIP
        '__main__.func1'

        >>> get_func_name(Class.func2)
        'Class.func2'

        >>> get_func_name(func3)
        'partial(...)'

        >>> get_func_name(func3, repr_fallback=True)
        'functools.partial(<built-in function len>)'


    Parameters
    ----------
    func : :class:`~collections.abc.Callable`
        A callable object.
    prepend_module : :class:`bool`
        If :data:`True` prepend the objects module (:attr:`~definition.__module__`),
        if available, to the to-be returned string.
    repr_fallback : :class:`bool`
        By default, when the passed function has neither a :attr:`~definition.__qualname__` or
        :attr:`~definition.__name__` attribute the (called) name of the functions class is returned.
        If :data:`True` then use :func:`repr` instead.

    Returns
    -------
    :class:`str`
        A string representation of the name of **func**.

    """
    try:
        name: str = getattr(func, '__qualname__', func.__name__)
    except AttributeError as ex:
        if not callable(func):
            raise TypeError("'func' expected a callable; "
                            f"observed type: {func.__class__.__name__!r}") from ex
        if repr_fallback:
            name = repr(func)
        else:
            name = f'{func.__class__.__name__}(...)'

    if prepend_module:
        try:
            return f'{func.__module__}.{name}'
        except AttributeError:
            pass
    return name


def group_by_values(iterable: Iterable[Tuple[_VT, _KT]]) -> Dict[_KT, List[_VT]]:
    """Take an iterable, yielding 2-tuples, and group all first elements by the second.

    Examples
    --------
    .. code:: python

        >>> str_list: list = ['a', 'a', 'a', 'a', 'a', 'b', 'b', 'b', 'c']
        >>> iterator = enumerate(str_list, 1)

        >>> new_dict: dict = group_by_values(iterator)
        >>> print(new_dict)
        {'a': [1, 2, 3, 4, 5], 'b': [6, 7, 8], 'c': [9]}

    Parameters
    ----------
    iterable : :class:`Iterable[Tuple[VT, KT]]<typing.Iterable>`
        An iterable yielding 2 elements upon iteration
        (*e.g.* :meth:`dict.items` or :func:`enumerate`).
        The second element must be :class:`Hashable<collections.abc.Hashable>` and will be used
        as key in the to-be returned mapping.

    Returns
    -------
    :class:`Dict[KT, List[VT]]<typing.Dict>`
        A grouped dictionary.

    """
    ret = {}
    list_append: Dict[_KT, Callable[[_VT], None]] = {}
    for value, key in iterable:
        try:
            list_append[key](value)
        except KeyError:
            ret[key] = [value]
            list_append[key] = ret[key].append
    return ret


def set_docstring(docstring: None | str) -> Callable[[_FT], _FT]:
    """A decorator for assigning docstrings.

    Examples
    --------
    .. code:: python

        >>> from nanoutils import set_docstring

        >>> number = "#10"

        >>> @set_docstring(f"Fancy docstring {number}.")
        ... def func():
        ...     pass

        >>> print(func.__doc__)
        Fancy docstring #10.

    Parameters
    ----------
    docstring : :class:`str`, optional
        The to-be assigned docstring.

    """
    def wrapper(func: _FT) -> _FT:
        func.__doc__ = docstring
        return func
    return wrapper


def get_importable(string: str, validate: None | Callable[[_T], bool] = None) -> _T:
    """Get an importable object.

    Examples
    --------
    .. code:: python

        >>> from inspect import isclass
        >>> from nanoutils import get_importable

        >>> dict_type = get_importable('builtins.dict', validate=isclass)
        >>> print(dict_type)
        <class 'dict'>

    Parameters
    ----------
    string : :class:`str`
        A string representing an importable object.
        Note that the string *must* contain the object's module.
    validate : :data:`~typing.Callable`, optional
        A callable for validating the imported object.
        Will raise a :exc:`RuntimeError` if its output evaluates to ``False``.

    Returns
    -------
    :class:`object`
        The imported object

    """
    try:
        head, *tail = string.split('.')
    except (AttributeError, TypeError) as ex:
        raise TypeError("'string' expected a str; observed type: "
                        f"{string.__class__.__name__!r}") from ex

    ret: _T = importlib.import_module(head)  # type: ignore
    for name in tail:
        ret = getattr(ret, name)

    if validate is None:
        return ret
    elif not validate(ret):
        val_str = get_func_name(validate) + '()'
        raise RuntimeError(f'Passing {ret!r} to {val_str} failed to return True')
    return ret


@overload
def split_dict(
    dct: MutableMapping[_KT, _VT],
    preserve_order: bool = ...,
    *,
    keep_keys: Iterable[_KT],
) -> Dict[_KT, _VT]:
    ...
@overload  # noqa: E302
def split_dict(
    dct: MutableMapping[_KT, _VT],
    preserve_order: bool = ...,
    *,
    disgard_keys: Iterable[_KT],
) -> Dict[_KT, _VT]:
    ...
def split_dict(dct, preserve_order=False, *, keep_keys=None, disgard_keys=None):  # noqa: E302,E501
    r"""Pop all items from **dct** which are in not in **keep_keys** and use them to construct a new dictionary.

    Note that, by popping its keys, the passed **dct** will also be modified inplace.

    Examples
    --------
    .. code:: python

        >>> from nanoutils import split_dict

        >>> dict1 = {1: 'a', 2: 'b', 3: 'c', 4: 'd'}
        >>> dict2 = split_dict(dict1, keep_keys={1, 2})

        >>> print(dict1, dict2, sep='\n')
        {1: 'a', 2: 'b'}
        {3: 'c', 4: 'd'}

        >>> dict3 = split_dict(dict1, disgard_keys={1, 2})
        >>> print(dict1, dict3, sep='\n')
        {}
        {1: 'a', 2: 'b'}

    Parameters
    ----------
    dct : :class:`MutableMapping[KT, VT]<typing.MutableMapping>`
        A mutable mapping.
    preserve_order : :class:`bool`
        If :data:`True`, preserve the order of the items in **dct**.
        Note that :code:`preserve_order = False` is generally faster.
    keep_keys : :class:`Iterable[KT]<typing.Iterable>`, keyword-only
        An iterable with keys that should remain in **dct**.
        Note that **keep_keys** and **disgard_keys** are mutually exclusive.
    disgard_keys : :class:`Iterable[KT]<typing.Iterable>`, keyword-only
        An iterable with keys that should be removed from **dct**.
        Note that **disgard_keys** and **keep_keys** are mutually exclusive.

    Returns
    -------
    :class:`Dict[KT, VT]<typing.Dict>`
        A new dictionaries with all key/value pairs from **dct** not specified in **keep_keys**.

    """  # noqa: E501
    if keep_keys is disgard_keys is None:
        raise TypeError("'keep_keys' and 'disgard_keys' cannot both be unspecified")
    elif keep_keys is None:
        iterable = _disgard_keys(dct, disgard_keys, preserve_order)
    elif disgard_keys is None:
        iterable = _keep_keys(dct, keep_keys, preserve_order)
    else:
        raise TypeError("'keep_keys' and 'disgard_keys' cannot both be specified")

    return {k: dct.pop(k) for k in iterable}


def _keep_keys(
    dct: Mapping[_KT, _VT],
    keep_keys: Iterable[_KT],
    preserve_order: bool = False,
) -> Collection[_KT]:
    """A helper function for :func:`split_dict`; used when :code:`keep_keys is not None`."""
    if preserve_order:
        return [k for k in dct if k not in keep_keys]
    else:
        try:
            return dct.keys() - keep_keys  # type: ignore
        except TypeError:
            return set(dct.keys()).difference(keep_keys)


def _disgard_keys(
    dct: Mapping[_KT, _VT],
    keep_keys: Iterable[_KT],
    preserve_order: bool = False,
) -> Collection[_KT]:
    """A helper function for :func:`split_dict`; used when :code:`disgard_keys is not None`."""
    if preserve_order:
        return [k for k in dct if k in keep_keys]
    else:
        try:
            return dct.keys() & keep_keys  # type: ignore
        except TypeError:
            return set(dct.keys()).intersection(keep_keys)


def raise_if(exception: None | BaseException) -> Callable[[_FT], _FT]:
    """A decorator which raises the passed exception whenever calling the decorated function.

    If **exception** is :data:`None` then the decorated function will be called as usual.

    Examples
    --------
    .. code:: python

        >>> from nanoutils import raise_if

        >>> ex1 = None
        >>> ex2 = TypeError("This is an exception")

        >>> @raise_if(ex1)
        ... def func1() -> bool:
        ...     return True

        >>> @raise_if(ex2)
        ... def func2() -> bool:
        ...     return True

        >>> func1()
        True

        >>> func2()
        Traceback (most recent call last):
          ...
        TypeError: This is an exception


    Parameters
    ----------
    exception : :exc:`BaseException`, optional
        An exception.
        If :data:`None` is passed then the decorated function will be called as usual.

    See Also
    --------
    :func:`nanoutils.ignore_if`
        A decorator which, if an exception is passed, ignores calls to the decorated function.

    """
    if exception is None:
        def decorator1(func: _FT) -> _FT:
            return func
        return decorator1

    elif isinstance(exception, BaseException):
        def decorator2(func: _FT) -> _FT:
            @wraps(func)
            def wrapper(*args, **kwargs):
                raise exception
            return wrapper  # type: ignore[return-value]
        return decorator2

    else:
        raise TypeError(f"{exception.__class__.__name__!r}")


def ignore_if(exception: None | BaseException, warn: bool = True) -> Callable[[_FT], _FT]:
    """A decorator which, if an exception is passed, ignores calls to the decorated function.

    If **exception** is :data:`None` then the decorated function will be called as usual.

    Examples
    --------
    .. code:: python

        >>> import warnings
        >>> from nanoutils import ignore_if

        >>> ex1 = None
        >>> ex2 = TypeError("This is an exception")

        >>> @ignore_if(ex1)
        ... def func1() -> bool:
        ...     return True

        >>> @ignore_if(ex2)
        ... def func2() -> bool:
        ...     return True

        >>> func1()
        True

        >>> func2()

        # Catch the warning and a raise it as an exception
        >>> with warnings.catch_warnings():
        ...     warnings.simplefilter("error", UserWarning)
        ...     func2()
        Traceback (most recent call last):
          ...
        UserWarning: Skipping call to func2()

    Parameters
    ----------
    exception : :exc:`BaseException`, optional
        An exception.
        If :data:`None` is passed then the decorated function will be called as usual.
    warn : :class:`bool`
        If :data:`True` issue a :exc:`UserWarning` whenever calling the decorated function

    See Also
    --------
    :func:`nanoutils.raise_if`
        A decorator which raises the passed exception whenever calling the decorated function.

    """
    _WARN = warn

    if exception is None:
        def decorator1(func: _FT) -> _FT:
            return func
        return decorator1

    elif isinstance(exception, BaseException):
        def decorator2(func: _FT) -> _FT:
            msg = f"Skipping call to {get_func_name(func)}()"

            @wraps(func)
            def wrapper(*args: Any, **kwargs: Any) -> None:
                if _WARN:
                    exc = UserWarning(msg)
                    exc.__cause__ = exception
                    warnings.warn(exc)
                    return None
            return wrapper  # type: ignore[return-value]
        return decorator2

    else:
        raise TypeError(f"{exception.__class__.__name__!r}")


_PATTERN = re.compile("([0-9]+).([0-9]+).([0-9]+)")


class VersionInfo(NamedTuple):
    """A :func:`~collections.namedtuple` representing the version of a package.

    Examples
    --------
    .. code:: python

        >>> from nanoutils import VersionInfo

        >>> version = '0.8.2'
        >>> VersionInfo.from_str(version)
        VersionInfo(major=0, minor=8, micro=2)

    """

    #: :class:`int`: The semantic_ major version.
    major: int

    #: :class:`int`: The semantic_ minor version.
    minor: int

    #: :class:`int`: The semantic_ micro version (a.k.a. :attr:`VersionInfo.patch`).
    micro: int

    @property
    def patch(self) -> int:
        """:class:`int`: An alias for :attr:`VersionInfo.micro`."""
        return self.micro

    @classmethod
    def from_str(
        cls,
        version: str,
        *,
        fullmatch: bool = True,
    ) -> VersionInfo:
        """Construct a :class:`VersionInfo` from a string.

        Parameters
        ----------
        version : :class:`str`
            A string representation of a version (*e.g.* :code:`version = "0.8.2"`).
            The string should contain three ``"."`` separated integers, respectively,
            representing the major, minor and micro/patch versions.
        fullmatch : :class:`bool`
            Whether the version-string must consist exclusivelly of three
            period-separated integers, or if a substring is also allowed.

        Returns
        -------
        :class:`nanoutils.VersionInfo`
            A new VersionInfo instance.

        """
        match = _PATTERN.fullmatch(version) if fullmatch else _PATTERN.match(version)
        if match is None:
            raise ValueError(f"Failed to parse {version!r}")
        return cls._make(int(i) for i in match.groups())


def _get_directive(
    obj: object,
    name: str,
    decorators: Container[str] = EMPTY_CONTAINER,
) -> str:
    """A helper function for :func:`~nanoutils.construct_api_doc`."""
    if isinstance(obj, type):
        if issubclass(obj, BaseException):
            return f'.. autoexception:: {name}'
        return f'.. autoclass:: {name}\n    :members:'

    elif callable(obj):
        if name in decorators:
            return f'.. autodecorator:: {name}'
        return f'.. autofunction:: {name}'

    elif isinstance(obj, ModuleType):
        return f'.. automodule:: {name}'

    else:
        return f'.. autodata:: {name}'


def construct_api_doc(
    glob_dict: Mapping[str, object],
    decorators: Container[str] = EMPTY_CONTAINER,
) -> str:
    '''Format a **Nano-Utils** module-level docstring.

    Examples
    --------
    .. code:: python

        >>> __doc__ = """
        ... Index
        ... -----
        ... .. autosummary::
        ... {autosummary}
        ...
        ... API
        ... ---
        ... {autofunction}
        ...
        ... """

        >>> from nanoutils import construct_api_doc

        >>> __all__ = ['obj', 'func', 'Class']

        >>> obj = ...

        >>> def func(obj: object) -> None:
        ...     pass

        >>> class Class(object):
        ...     pass

        >>> doc = construct_api_doc(locals())
        >>> print(doc)
        <BLANKLINE>
        Index
        -----
        .. autosummary::
            obj
            func
            Class
        <BLANKLINE>
        API
        ---
        .. autodata:: obj
        .. autofunction:: func
        .. autoclass:: Class
            :members:
        <BLANKLINE>
        <BLANKLINE>

    Parameters
    ----------
    glob_dict : :class:`Mapping[str, object]<typing.Mapping>`
        A mapping containg a module-level namespace.
        Note that the mapping *must* contain the ``"__doc__"`` and ``"__all__"`` keys.

    decorators : :class:`Container[str]<typing.Container>`
        A container with the names of all decorators.
        If not specified, all functions will use the Sphinx ``autofunction`` domain.

    Returns
    -------
    :class:`str`
        The formatted string.

    '''
    __doc__ = cast(str, glob_dict['__doc__'])
    __all__ = cast(List[str], glob_dict['__all__'])

    return __doc__.format(
        autosummary='\n'.join(f'    {i}' for i in __all__),
        autofunction='\n'.join(_get_directive(glob_dict[i], i, decorators) for i in __all__)
    )


def _null_func(obj: _T) -> _T:
    """Return the passed object."""
    return obj


def positional_only(func: _FT) -> _FT:
    """A decorator for converting mangled parameters to positional-only.

    Sets the ``__signature__`` attribute of the decorated function.

    Examples
    --------
    .. code-block:: python

        >>> from nanoutils import positional_only
        >>> import inspect

        >>> def func1(__a, b=None):
        ...     pass

        >>> print(inspect.signature(func1))
        (__a, b=None)

        >>> @positional_only
        ... def func2(__a, b=None):
        ...     pass

        >>> print(inspect.signature(func2))
        (a, /, b=None)

    Parameters
    ----------
    func : :class:`~collections.abc.Callable`
        The to-be decorated function whose ``__signature__`` attribute will be added or updated.

    """
    # Check if any name mangling has occured
    cls_name, period, _ = getattr(func, "__qualname__", "").partition(".")
    prefix = "__" if not period else f"_{cls_name}__"
    offset = len(prefix)

    # Identify the if `__`-prefixed parameters
    sgn = inspect.signature(func)
    pos_only_list = []
    for j, name in enumerate(sgn.parameters):
        if name.startswith(prefix):
            pos_only_list.append(j)

    if not pos_only_list:
        func.__signature__ = sgn  # type: ignore[attr-defined]
        return func
    else:
        j = pos_only_list[-1]

    # Unmangle parameters and convert them to positional-only
    prm_list = []
    for i, (name, prm) in enumerate(sgn.parameters.items()):
        if i <= j:
            if name.startswith(prefix):
                name = name[offset:]
            prm = prm.replace(name=name, kind=inspect.Parameter.POSITIONAL_ONLY)
        prm_list.append(prm)

    func.__signature__ = inspect.Signature(   # type: ignore[attr-defined]
        parameters=prm_list,
        return_annotation=sgn.return_annotation,
    )
    func.__annotations__ = {
        (k[offset:] if k.startswith(prefix) else k): v for k, v in func.__annotations__.items()
    }
    return func


# Move to the end to reduce the risk of circular imports
from ._partial import PartialPrepend
from ._set_attr import SetAttr
from ._seq_view import SequenceView
from ._catch_err import CatchErrors
from ._lazy_import import LazyImporter, MutableLazyImporter
from ._user_dict import UserMapping, MutableUserMapping

__doc__ = construct_api_doc(
    globals(),
    decorators={'set_docstring', 'raise_if', 'ignore_if', 'positional_only'},
)



"""A private module for containg the :class:`nanoutils.SequenceView` class.

Notes
-----
:class:`~nanoutils.SequenceView` should be imported from
either :mod:`nanoutils` or :mod:`nanoutils.utils`.

"""

from __future__ import annotations

import sys
import pprint
import textwrap
from inspect import Signature, Parameter
from typing import (
    Any,
    TypeVar,
    Sequence,
    NoReturn,
    overload,
    Iterator,
    Dict,
    Callable,
    ClassVar,
    TYPE_CHECKING,
)

from .typing_utils import TypedDict, SupportsIndex

if sys.version_info >= (3, 8):
    class _PPrintDict(TypedDict, total=False):
        indent: int
        width: int
        depth: None | int
        compact: bool
        sort_dicts: bool
else:
    class _PPrintDict(TypedDict, total=False):
        indent: int
        width: int
        depth: None | int
        compact: bool

_T_co = TypeVar("_T_co", covariant=True)
_FT = TypeVar("_FT", bound=Callable[..., Any])
_SVT = TypeVar("_SVT", bound="SequenceView[Any]")

__all__ = ["SequenceView"]


def _set_signature(signature: Signature) -> Callable[[_FT], _FT]:
    """Set the ``__signature__`` and ``__annotations__`` of the decorated function."""
    def _decorate(func: _FT) -> _FT:
        func.__signature__ = signature  # type: ignore[attr-defined]
        func.__annotations__ = {k: v.annotation for k, v in signature.parameters.items() if
                                v.annotation is not Signature.empty}
        if signature.return_annotation is not Signature.empty:
            func.__annotations__["return"] = signature.return_annotation
        return func
    return _decorate


_INDEX_SIGNATURE = Signature([
    Parameter("self", Parameter.POSITIONAL_ONLY),
    Parameter("value", Parameter.POSITIONAL_ONLY, annotation=Any),
    Parameter("start", Parameter.POSITIONAL_ONLY, default=0, annotation=SupportsIndex),
    Parameter("stop", Parameter.POSITIONAL_ONLY, default=sys.maxsize, annotation=SupportsIndex),
], return_annotation=int)

_COUNT_SIGNATURE = Signature([
    Parameter("self", Parameter.POSITIONAL_ONLY),
    Parameter("value", Parameter.POSITIONAL_ONLY, annotation=Any),
], return_annotation=int)


class SequenceView(Sequence[_T_co]):
    """A read-only view of an underlying :class:`~collections.abc.Sequence`.

    Examples
    --------
    .. code-block:: python

        >>> from nanoutils import SequenceView

        >>> lst = [1, 2, 3]
        >>> view = SequenceView(lst)
        >>> print(view)
        SequenceView([1, 2, 3])

        >>> lst.append(4)
        >>> print(view)
        SequenceView([1, 2, 3, 4])

        >>> print(len(view))
        4

        >>> del view[0]
        Traceback (most recent call last):
            ...
        TypeError: 'SequenceView' object doesn't support item deletion

    """

    __slots__ = ("__weakref__", "_seq")

    #: A class variable containing a dictionary with keyword arguments for :func:`pprint.pformat`.
    pprint_kwargs: ClassVar[_PPrintDict] = NotImplemented

    def __init__(self, sequence: Sequence[_T_co]) -> None:
        r"""Initialize this instance.

        Parameters
        ----------
        sequence : :class:`~collections.abc.Sequence`
            The to-be wrapped sequence.

        """
        self._seq: Sequence[_T_co] = sequence

    def __init_subclass__(cls) -> None:
        """Attach a unique :attr:`~SequenceView.pprint_kwargs` dict to each subclass."""
        super().__init_subclass__()
        width = 80 - 1 - len(cls.__name__)
        if cls.pprint_kwargs is NotImplemented:
            dct: _PPrintDict = {"compact": True, "width": width}
            if sys.version_info >= (3, 8):
                dct["sort_dicts"] = False
        else:
            dct = cls.pprint_kwargs.copy()
            dct["width"] = width
        cls.pprint_kwargs = dct

    def __hash__(self) -> int:
        """Implement :func:`hash(sellf) <hash>`."""
        if isinstance(self._seq, SequenceView):
            # delegate to the underlying `SequenceView` instance
            return hash(self._seq)
        else:
            return id(self._seq)

    def __reduce__(self) -> NoReturn:
        """Helper method for :mod:`pickle`."""
        raise TypeError(f"cannot pickle {type(self).__name__!r} object")

    def __copy__(self: _SVT) -> _SVT:
        """Return a shallow copy of this instance."""
        return self  # `self` is immutable

    def __deepcopy__(self: _SVT, memo: None | Dict[int, Any] = None) -> _SVT:
        """Return a deep copy of this instance."""
        return self  # `self` is immutable

    @overload
    def __getitem__(self, key: SupportsIndex) -> _T_co:
        ...
    @overload  # noqa: E301
    def __getitem__(self: _SVT, key: slice) -> _SVT:
        ...
    def __getitem__(self: _SVT, key: SupportsIndex | slice) -> _SVT | _T_co:  # noqa: E301
        """Implement :meth:`self[key] <object.__getitem__>`."""
        if isinstance(key, slice):
            cls = type(self)
            return cls(self._seq[key])
        return self._seq[key]  # type: ignore

    if TYPE_CHECKING:
        def index(self, __value: Any, __start: SupportsIndex = ..., __stop: SupportsIndex = ...) -> int:  # noqa: D102,E501
            ...

        def count(self, __value: Any) -> int:  # noqa: D102
            ...
    else:
        @_set_signature(_INDEX_SIGNATURE)
        def index(self, *args, **kwargs):
            """Return the first index of **value**."""
            return self._seq.index(*args, **kwargs)

        @_set_signature(_COUNT_SIGNATURE)
        def count(self, *args, **kwargs):
            """Return the number of times **value** occurs in the instance."""
            return self._seq.count(*args, **kwargs)

    def __len__(self):
        """Implement :func:`len(self) <len>`."""
        return len(self._seq)

    def __contains__(self, x: object) -> bool:
        """Implement :meth:`x in self <object.__contains__>`."""
        return x in self._seq

    def __iter__(self) -> Iterator[_T_co]:
        """Implement :func:`iter(self) <iter>`."""
        return iter(self._seq)

    def __reversed__(self) -> Iterator[_T_co]:
        """Implement :func:`reversed(self) <reversed>`."""
        return reversed(self._seq)

    def __repr__(self) -> str:
        """Implement :func:`repr(self) <repr>`."""
        cls = type(self)
        name = cls.__name__
        offset = len(name) + 1

        seq = pprint.pformat(self._seq, **cls.pprint_kwargs)
        seq_indent = textwrap.indent(seq, " " * offset)[offset:]
        return f"{name}({seq_indent})"

    def __eq__(self, other: object) -> bool:
        """Implement :meth:`x == self <object.__eq__>`."""
        return self._seq == other


# Ensure that `__init_subclass__` is called not just for `SequenceView` subclasses,
# but also the superclass itself.
SequenceView.__init_subclass__()
